/*
 *  linux/init/main.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  GK 2/5/95  -  Changed to support mounting root fs via NFS
 *  Added initrd & change_root: Werner Almesberger & Hans Lermen, Feb '96
 *  Moan early if gcc is old, avoiding bogus kernels - Paul Gortmaker, May '96
 *  Simplified starting of init:  Michael A. Griffith <grif@acm.org> 
 */

#define __KERNEL_SYSCALLS__

#include <linux/config.h>
/////////////// Add by Vic Lee, ASUS ///////////////////
#define EXPORT_SYMTAB
#ifdef CONFIG_BRECIS
#include <linux/module.h>
#endif
/////////////// End of Vic Lee, ASUS ///////////////////
#include <linux/proc_fs.h>
#include <linux/devfs_fs_kernel.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <linux/delay.h>
#include <linux/utsname.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/smp_lock.h>
#include <linux/blk.h>
#include <linux/hdreg.h>
#include <linux/iobuf.h>
#include <linux/bootmem.h>
#include <linux/file.h>
#include <linux/tty.h>
#include <asm/uaccess.h>

#include <asm/io.h>
#include <asm/bugs.h>

#if defined(CONFIG_ARCH_S390)
#include <asm/s390mach.h>
#include <asm/ccwcache.h>
#endif

#ifdef CONFIG_ACPI
#include <linux/acpi.h>
#endif

#ifdef CONFIG_PCI
#include <linux/pci.h>
#endif

#ifdef CONFIG_DIO
#include <linux/dio.h>
#endif

#ifdef CONFIG_ZORRO
#include <linux/zorro.h>
#endif

#ifdef CONFIG_MTRR
#  include <asm/mtrr.h>
#endif

#ifdef CONFIG_NUBUS
#include <linux/nubus.h>
#endif

#ifdef CONFIG_ISAPNP
#include <linux/isapnp.h>
#endif

#ifdef CONFIG_IRDA
extern int irda_proto_init(void);
extern int irda_device_init(void);
#endif

#ifdef CONFIG_X86_LOCAL_APIC
#include <asm/smp.h>
#endif

/*
 * Versions of gcc older than that listed below may actually compile
 * and link okay, but the end product can have subtle run time bugs.
 * To avoid associated bogus bug reports, we flatly refuse to compile
 * with a gcc that is known to be too old from the very beginning.
 */
#if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 91)
#if !defined(CONFIG_V850E) && !defined(CONFIG_NIOS)
#error Sorry, your GCC is too old. It builds incorrect kernels.
#else
#warning Sorry, your GCC is too old. It builds incorrect kernels.
#endif
#endif

extern char _stext, _etext;
extern char *linux_banner;

static int init(void *);

extern void init_IRQ(void);
extern void init_modules(void);
extern void sock_init(void);
extern void fork_init(unsigned long);
extern void mca_init(void);
extern void sbus_init(void);
extern void ppc_init(void);
extern void sysctl_init(void);
extern void signals_init(void);
extern int init_pcmcia_ds(void);

extern void free_initmem(void);

#ifdef CONFIG_TC
extern void tc_init(void);
#endif

extern void ecard_init(void);

#if defined(CONFIG_SYSVIPC)
extern void ipc_init(void);
#endif

#if defined(CONFIG_BRECIS) && defined(CONFIG_GMON_PROF)
extern void __monstartup(unsigned int *lowpc, unsigned int *highp);
extern void __monstartup_after_bootmem(void);
extern int _ftext, _end;
#endif	/* CONFIG_BRECIS && CONFIG_GMON_PROF */

/*
 * Boot command-line arguments
 */
#define MAX_INIT_ARGS 8
#define MAX_INIT_ENVS 8

extern void time_init(void);
extern void softirq_init(void);

int rows, cols;

char *execute_command;

static char * argv_init[MAX_INIT_ARGS+2] = { "init", NULL, };
char * envp_init[MAX_INIT_ENVS+2] = { "HOME=/", "TERM=linux", NULL, };

static int __init profile_setup(char *str)
{
    int par;
    if (get_option(&str,&par)) prof_shift = par;
	return 1;
}

__setup("profile=", profile_setup);

static int __init checksetup(char *line)
{
	struct kernel_param *p;

	p = &__setup_start;
	do {
		int n = strlen(p->str);
		if (!strncmp(line,p->str,n)) {
			if (p->setup_func(line+n))
				return 1;
		}
		p++;
	} while (p < &__setup_end);
	return 0;
}

/* this should be approx 2 Bo*oMips to start (note initial shift), and will
   still work even if initially too large, it will just take slightly longer */
unsigned long loops_per_jiffy = (1<<12);

/* This is the number of bits of precision for the loops_per_jiffy.  Each
   bit takes on average 1.5/HZ seconds.  This (like the original) is a little
   better than 1% */
#define LPS_PREC 8

void __init calibrate_delay(void)
{
	unsigned long ticks, loopbit;
	int lps_precision = LPS_PREC;

#ifdef FIXED_BOGOMIPS
	int bogus;

/* FIXED_BOGOMIPS converted to __delay units.  */
#define FIXED_LOOPS_PER_JIFFY	(unsigned long)(FIXED_BOGOMIPS * (500000 / HZ))

/* The maximum error in FIXED_LOOPS_PER_JIFFY that we will tolerate.  */
#define FIXED_LPJ_TOLERANCE	(unsigned long)(FIXED_LOOPS_PER_JIFFY * 0.10)

	/* Make sure fixed delay - T% is zero ticks.  */
	ticks = jiffies;
	while (ticks == jiffies) /* Synchronize with start of tick */
		/* nothing */;
	ticks = jiffies;
	__delay(FIXED_LOOPS_PER_JIFFY - FIXED_LPJ_TOLERANCE);
	bogus = (jiffies != ticks);

	if (! bogus) {
		/* Make sure fixed delay + T% is one tick.  The delay here
		   is very short because we're actually continuing timing from
		   the tick synchronization above (we don't resynchronize).  */
		__delay(2 * FIXED_LPJ_TOLERANCE);
		bogus = (jiffies != ticks + 1);
	}
	
	if (! bogus) {
		/* Use the precomputed value.  */
		loops_per_jiffy = FIXED_LOOPS_PER_JIFFY;
		printk("Delay loop constant: %lu.%02lu BogoMIPS (precomputed)\n",
		       (unsigned long)FIXED_BOGOMIPS,
		       ((unsigned long)(FIXED_BOGOMIPS * 100)) % 100);
		return;
	} else {
		printk("Precomputed BogoMIPS value (%lu.%02lu) inaccurate!\n",
		       (unsigned long)FIXED_BOGOMIPS,
		       ((unsigned long)(FIXED_BOGOMIPS * 100)) % 100);
		/* ... and fall through to normal bogomips calculation.  */
	}
#endif /* FIXED_BOGOMIPS */

	loops_per_jiffy = (1<<12);

	printk("Calibrating delay loop... ");
	while (loops_per_jiffy <<= 1) {
		/* wait for "start of" clock tick */
		ticks = jiffies;
		while (ticks == jiffies)
			/* nothing */;
		/* Go .. */
		ticks = jiffies;
		__delay(loops_per_jiffy);
		ticks = jiffies - ticks;
		if (ticks)
			break;
	}

/* Do a binary approximation to get loops_per_jiffy set to equal one clock
   (up to lps_precision bits) */
	loops_per_jiffy >>= 1;
	loopbit = loops_per_jiffy;
	while ( lps_precision-- && (loopbit >>= 1) ) {
		loops_per_jiffy |= loopbit;
		ticks = jiffies;
		while (ticks == jiffies);
		ticks = jiffies;
		__delay(loops_per_jiffy);
		if (jiffies != ticks)	/* longer than 1 tick */
			loops_per_jiffy &= ~loopbit;
	}

/* Round the value and print it */	
	printk("%lu.%02lu BogoMIPS\n",
		loops_per_jiffy/(500000/HZ),
		(loops_per_jiffy/(5000/HZ)) % 100);
}

static int __init debug_kernel(char *str)
{
	if (*str)
		return 0;
	console_loglevel = 10;
	return 1;
}

static int __init quiet_kernel(char *str)
{
	if (*str)
		return 0;
	console_loglevel = 4;
	return 1;
}

__setup("debug", debug_kernel);
__setup("quiet", quiet_kernel);

/*
 * This is a simple kernel command line parsing function: it parses
 * the command line, and fills in the arguments/environment to init
 * as appropriate. Any cmd-line option is taken to be an environment
 * variable if it contains the character '='.
 *
 * This routine also checks for options meant for the kernel.
 * These options are not given to init - they are for internal kernel use only.
 */
static void __init parse_options(char *line)
{
	char *next,*quote;
	int args, envs;

	if (!*line)
		return;
	args = 0;
	envs = 1;	/* TERM is set to 'linux' by default */
	next = line;
	while ((line = next) != NULL) {
                quote = strchr(line,'"');
                next = strchr(line, ' ');
                while (next != NULL && quote != NULL && quote < next) {
                        /* we found a left quote before the next blank
                         * now we have to find the matching right quote
                         */
                        next = strchr(quote+1, '"');
                        if (next != NULL) {
                                quote = strchr(next+1, '"');
                                next = strchr(next+1, ' ');
                        }
                }
                if (next != NULL)
                        *next++ = 0;
		if (!strncmp(line,"init=",5)) {
			line += 5;
			execute_command = line;
			/* In case LILO is going to boot us with default command line,
			 * it prepends "auto" before the whole cmdline which makes
			 * the shell think it should execute a script with such name.
			 * So we ignore all arguments entered _before_ init=... [MJ]
			 */
			args = 0;
			continue;
		}
		if (checksetup(line))
			continue;
		
		/*
		 * Then check if it's an environment variable or
		 * an option.
		 */
		if (strchr(line,'=')) {
			if (envs >= MAX_INIT_ENVS)
				break;
			envp_init[++envs] = line;
		} else {
			if (args >= MAX_INIT_ARGS)
				break;
			if (*line)
				argv_init[++args] = line;
		}
	}
	argv_init[args+1] = NULL;
	envp_init[envs+1] = NULL;
}


extern void setup_arch(char **);
extern void cpu_idle(void);

unsigned long wait_init_idle;

#ifndef CONFIG_SMP

#ifdef CONFIG_X86_LOCAL_APIC
static void __init smp_init(void)
{
	APIC_init_uniprocessor();
}
#else
#define smp_init()	do { } while (0)
#endif

#else


/* Called by boot processor to activate the rest. */
static void __init smp_init(void)
{
	/* Get other processors into their bootup holding patterns. */
	smp_boot_cpus();
	wait_init_idle = cpu_online_map;
	clear_bit(current->processor, &wait_init_idle); /* Don't wait on me! */

	smp_threads_ready=1;
	smp_commence();

	/* Wait for the other cpus to set up their idle processes */
	printk("Waiting on wait_init_idle (map = 0x%lx)\n", wait_init_idle);
	while (wait_init_idle) {
		cpu_relax();
		barrier();
	}
	printk("All processors have done init_idle\n");
}

#endif

/*
 * We need to finalize in a non-__init function or else race conditions
 * between the root thread and the init thread may cause start_kernel to
 * be reaped by free_initmem before the root thread has proceeded to
 * cpu_idle.
 */

static void rest_init(void)
{
	kernel_thread(init, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGNAL);
	unlock_kernel();
	current->need_resched = 1;
 	cpu_idle();
} 

/*
 *	Activate the first processor.
 */
#ifdef CONFIG_BRECIS
#if defined(CONFIG_CHECKSTACK) || defined(CONFIG_CHECKLOWMEMORYZERO)
extern int __console_inited__;
#endif	/* CONFIG_CHECKSTACK || CONFIG_CHECKLOWMEMORYZERO */
#endif	/* CONFIG_BRECIS */

asmlinkage void __init start_kernel(void)
{
	char * command_line;
	extern char saved_command_line[];

/*
 * Interrupts are still disabled. Do necessary setups, then
 * enable them
 */
	lock_kernel();
	printk(linux_banner);
#if defined(CONFIG_BRECIS) && defined(CONFIG_GMON_PROF)
	printk("Starting __monstartup for CONFIG_GMON_PROF\n");
	__monstartup(&_ftext, &_end);
#endif	/* CONFIG_BRECIS && CONFIG_GMON_PROF */
	setup_arch(&command_line);
#if defined(CONFIG_BRECIS) && defined(CONFIG_GMON_PROF)
	/* this was moved out of __monstartup() and placed into a new function,
	 * with the hope that correcting the {init,alloc}_bootmem call sequence
	 * would allow our kernel to boot. AFAIK, this has no adverse effect
	 * on the rest of the kernel, other than missing a few profiling
	 * opportunities within setup_arch(). */
	printk("Finishing GMON_PROF initialization\n");
	__monstartup_after_bootmem();
#endif /* CONFIG_BRECIS && CONFIG_GMON_PROF */
	printk("Kernel command line: %s\n", saved_command_line);
	parse_options(command_line);
	trap_init();
	init_IRQ();
	sched_init();
	softirq_init();
	time_init();

#ifdef CONFIG_BRECIS
#if defined(CONFIG_CHECKSTACK) || defined(CONFIG_CHECKLOWMEMORYZERO)
__console_inited__ = 1;
#endif  /* CONFIG_CHECKSTACK || CONFIG_CHECKLOWMEMORYZERO */
#endif	/* CONFIG_BRECIS */
	/*
	 * HACK ALERT! This is early. We're enabling the console before
	 * we've done PCI setups etc, and console_init() must be aware of
	 * this. But we do want output early, in case something goes wrong.
	 */
	console_init();
#ifdef CONFIG_MODULES
	init_modules();
#endif
	if (prof_shift) {
		unsigned int size;
		/* only text is profiled */
		prof_len = (unsigned long) &_etext - (unsigned long) &_stext;
		prof_len >>= prof_shift;
		
		size = prof_len * sizeof(unsigned int) + PAGE_SIZE-1;
		prof_buffer = (unsigned int *) alloc_bootmem(size);
	}

	kmem_cache_init();
	sti();
	calibrate_delay();
#ifdef CONFIG_BLK_DEV_INITRD
	if (initrd_start && !initrd_below_start_ok &&
			initrd_start < min_low_pfn << PAGE_SHIFT) {
		printk(KERN_CRIT "initrd overwritten (0x%08lx < 0x%08lx) - "
		    "disabling it.\n",initrd_start,min_low_pfn << PAGE_SHIFT);
		initrd_start = 0;
	}
#endif
	mem_init();
	kmem_cache_sizes_init();
	pgtable_cache_init();

	/*
	 * For architectures that have highmem, num_mappedpages represents
	 * the amount of memory the kernel can use.  For other architectures
	 * it's the same as the total pages.  We need both numbers because
	 * some subsystems need to initialize based on how much memory the
	 * kernel can use.
	 */
	if (num_mappedpages == 0)
		num_mappedpages = num_physpages;
  
	fork_init(num_mappedpages);
	proc_caches_init();
	vfs_caches_init(num_physpages);
	buffer_init(num_physpages);
	page_cache_init(num_physpages);
#if defined(CONFIG_ARCH_S390)
	ccwcache_init();
#endif
	signals_init();
#ifdef CONFIG_PROC_FS
	proc_root_init();
#endif
	check_bugs();
	printk("POSIX conformance testing by UNIFIX\n");

	/* 
	 *	We count on the initial thread going ok 
	 *	Like idlers init is an unlocked kernel thread, which will
	 *	make syscalls (and thus be locked).
	 */
	smp_init();
#if defined(CONFIG_SYSVIPC)
	ipc_init();
#endif
	rest_init();
}

struct task_struct *child_reaper = &init_task;

static void __init do_initcalls(void)
{
	initcall_t *call;

	call = &__initcall_start;
	do {
		(*call)();
		call++;
	} while (call < &__initcall_end);

	/* Make sure there is no pending stuff from the initcall sequence */
	flush_scheduled_tasks();
}

/*
 * Ok, the machine is now initialized. None of the devices
 * have been touched yet, but the CPU subsystem is up and
 * running, and memory and process management works.
 *
 * Now we can finally start doing some real work..
 */
static void __init do_basic_setup(void)
{

	/*
	 * Tell the world that we're going to be the grim
	 * reaper of innocent orphaned children.
	 *
	 * We don't want people to have to make incorrect
	 * assumptions about where in the task array this
	 * can be found.
	 */
	child_reaper = current;

#if defined(CONFIG_MTRR)	/* Do this after SMP initialization */
/*
 * We should probably create some architecture-dependent "fixup after
 * everything is up" style function where this would belong better
 * than in init/main.c..
 */
	mtrr_init();
#endif

#ifdef CONFIG_SYSCTL
	sysctl_init();
#endif

	/*
	 * Ok, at this point all CPU's should be initialized, so
	 * we can start looking into devices..
	 */
#if defined(CONFIG_ARCH_S390)
	s390_init_machine_check();
#endif
#ifdef CONFIG_ACPI_INTERPRETER
	acpi_init();
#endif
#ifdef CONFIG_PCI
	pci_init();
#endif
#ifdef CONFIG_SBUS
	sbus_init();
#endif
#if defined(CONFIG_PPC)
	ppc_init();
#endif
#ifdef CONFIG_MCA
	mca_init();
#endif
#ifdef CONFIG_ARCH_ACORN
	ecard_init();
#endif
#ifdef CONFIG_ZORRO
	zorro_init();
#endif
#ifdef CONFIG_DIO
	dio_init();
#endif
#ifdef CONFIG_NUBUS
	nubus_init();
#endif
#ifdef CONFIG_ISAPNP
	isapnp_init();
#endif
#ifdef CONFIG_TC
	tc_init();
#endif

	/* Networking initialization needs a process context */ 
	sock_init();

	start_context_thread();
	do_initcalls();

#ifdef CONFIG_IRDA
	irda_proto_init();
	irda_device_init(); /* Must be done after protocol initialization */
#endif
#ifdef CONFIG_PCMCIA
	init_pcmcia_ds();		/* Do this last */
#endif
}

static void run_init_process(char *init_filename)
{
	argv_init[0] = init_filename;
	execve(init_filename, argv_init, envp_init);
}

extern void prepare_namespace(void);
////////////////////////////////////////////////////////
/////////////// Add by Vic Lee, ASUS ///////////////////
////////////////////////////////////////////////////////
#define V_PROC_BUF_LEN 300
#include <linux/kernel.h>
#include <linux/reboot.h>
#include <linux/inet.h>
#ifdef CONFIG_BRECIS
#include <asm/brecis/BrecisSysRegs.h>
#endif
static struct proc_dir_entry *sysinfo_entry;
static char sysinfo_buf[V_PROC_BUF_LEN];

static unsigned char watch_dog_running=0;
static struct timer_list wd_timer;
unsigned char V_Factory_Reset;
unsigned char V_serial[21];
unsigned char V_bootVer[16];
#ifdef CONFIG_BRECIS
unsigned char V_WAN_LINK=0;
unsigned char V_WAN_RATE;
unsigned char V_WAN_DUPLEX;
unsigned char V_MAC1[6];
unsigned char V_MAC2[6];
#endif
unsigned long V_LAN_IP=0, V_LAN_MASK, V_LAN_NET;
unsigned char V_fsize=0;
int V_model;
int V_wanlink=0;
#ifdef CONFIG_BRECIS
void VV_SetWAN_link(int type);
#endif
unsigned short mysip_listen_port=5060;
unsigned short mysip_rtp_start=50000, mysip_rtp_end=60000;
unsigned short mysip_stun_port[5]={0};
#ifdef CONFIG_BRECIS
#define GPIO_15_10_DATA_REG (volatile unsigned *) (0xB840017C)
#define _74HC164_DATA_HIGHT (*GPIO_15_10_DATA_REG |= 0x00000008)
#define _74HC164_DATA_LOW (*GPIO_15_10_DATA_REG &= 0xFFFFFFF7)
#define _74HC164_CLK_HIGHT (*GPIO_15_10_DATA_REG |= 0x00000010)
#define _74HC164_CLK_LOW (*GPIO_15_10_DATA_REG &= 0xFFFFFFEF)
static unsigned char VV_74hc164_value;

static void wd_refresh(unsigned long data)
{
 if (watch_dog_running)
  {
   *WD_SVC_REG=0x5A5A5A5A;
   wd_timer.expires = jiffies + HZ * 10; // 10 seconds
   add_timer(&wd_timer);
  }
}

static void VV_74HC164_WriteOut(void)
// Notice, we must make sure the GPIO direction is output for 74HC164 data/clock  
{
 int i;
// clear the clock
 _74HC164_CLK_HIGHT;
 _74HC164_CLK_LOW;
// write eight bits
 for (i=7;i>=0;i--)
  {
   _74HC164_CLK_HIGHT;
   if (VV_74hc164_value & (1<<i)) // high
    _74HC164_DATA_HIGHT;
   else
    _74HC164_DATA_LOW;
   _74HC164_CLK_LOW;
  }
// add final high low
 _74HC164_CLK_HIGHT;
 _74HC164_CLK_LOW;
}

void VV_74HC164_SetByte(unsigned char val)
{
 VV_74hc164_value=val;
 VV_74HC164_WriteOut();
}

void VV_74HC164_SetBit(int bit, int val)
{
 if (val) // set
  VV_74hc164_value |= (1<<bit);
 else // clear
  VV_74hc164_value &= ~(1<<bit);
 VV_74HC164_WriteOut();
}

int VV_74HC164_GetBit(int bit)
{
 if (VV_74hc164_value & (1<<bit))
  return 1;
 else
  return 0;
//printk("VV_74HC164_GetBit:[%d]=>%d\n",bit,(VV_74hc164_value & (1<<bit))?1:0);
}
#endif


//extern unsigned char V_Factory_Reset;
//extern unsigned char V_serial[21];
//extern unsigned char V_bootVer[16];
//extern int V_model;

static ssize_t sysinfo_read(struct file *file, char *ubuf, size_t bytes, loff_t *pos)
{
 char buf[1536];
 int len=0,rlen;

 if(bytes < 0) return -EINVAL;

 if(bytes == 0) return 0;

 len+=sprintf(buf+len,"factory=%d\n", V_Factory_Reset ? 1: 0 );
 len+=sprintf(buf+len,"serial=%s\n", V_serial);
 len+=sprintf(buf+len,"model=%d\n", V_model);
 len+=sprintf(buf+len,"fsize=%u\n", V_fsize);
 len+=sprintf(buf+len,"MAC1=%02X:%02X:%02X:%02X:%02X:%02X\n", V_MAC1[0], V_MAC1[1], V_MAC1[2], V_MAC1[3], V_MAC1[4], V_MAC1[5]);
 len+=sprintf(buf+len,"MAC2=%02X:%02X:%02X:%02X:%02X:%02X\n", V_MAC2[0], V_MAC2[1], V_MAC2[2], V_MAC2[3], V_MAC2[4], V_MAC2[5]);
 len+=sprintf(buf+len,"wanlink=%d\n", V_wanlink);
 len+=sprintf(buf+len,"SIP=%u,%u,%u\n",mysip_listen_port,mysip_rtp_start,mysip_rtp_end);
 len+=sprintf(buf+len,"LANIP=%u.%u.%u.%u, MASK=%u.%u.%u.%u\n", NIPQUAD(V_LAN_IP), NIPQUAD(V_LAN_MASK));
 len+=sprintf(buf+len,"bver=%s\n",V_bootVer);
 if (mysip_stun_port[0])
  {
   int i;
   len+=sprintf(buf+len,"STUN=%u",mysip_stun_port[0]);
   for (i=1;(i<5)&&(mysip_stun_port[i]!=0);i++)
    len+=sprintf(buf+len,",%u",mysip_stun_port[i]);
   len+=sprintf(buf+len,"\n");
  }
#ifdef CONFIG_BRECIS
 len+=sprintf(buf+len,"WAN_NIC=%u,%u,%u\n", V_WAN_LINK, V_WAN_RATE, V_WAN_DUPLEX);
#endif

 if (*pos >= len) return 0;

 rlen=len-*pos;
 if (rlen > bytes)
  rlen=bytes;
 if (copy_to_user(ubuf, buf+*pos, rlen))
  return -EFAULT;
   
 *pos += rlen;
 return rlen;
}

ssize_t sysinfo_write(struct file *file, const char *buf, size_t bytes, loff_t *pos)
{
 int retval;

 if (bytes < 0) return -EINVAL;
 if (bytes == 0) return 0;

 if (*pos==0)
   sysinfo_buf[0]=0x00;

 if ((bytes+*pos) >= V_PROC_BUF_LEN) bytes = (V_PROC_BUF_LEN-1) - *pos;

 retval = copy_from_user(sysinfo_buf + *pos, buf, bytes);
 if (retval != 0)
  {
   printk("sysinfo_write: copy_from_user() failed!\n");
   return -EFAULT;
  }

 *pos += bytes;
 sysinfo_buf[*pos] = '\0';

 return bytes;
}

unsigned long vic_atoul(char *name)
{
 unsigned long val = 0;

 for (;; name++)
  {
   switch (*name)
    {
     case '0'...'9':
       val = 10*val+(*name-'0');
       break;
     default:
       return val;
    }
  }
}

#ifdef CONFIG_IP_NF_TARGET_TRIGGER
extern unsigned long vic_pt_timeout;
#endif

typedef struct
{
 int offset;
 int index;
} gpio_defs;

static gpio_defs gpio_cfg_arr[] = {\
{ 0, 0 },
{ 0, 1 },
{ 4, 0 },
{ 4, 1 },
{ 4, 2 },
{ 4, 3 },
{ 8, 0 },
{ 8, 1 },
{ 8, 2 },
{ 8, 3 },
{ 12, 0 },
{ 12, 1 },
{ 12, 2 },
{ 12, 3 },
{ 12, 4 },
{ 12, 5 } 
};

#define v_inl(a)		(*((volatile unsigned long *)(a)))
#define v_outl(a,v)	(*((volatile unsigned long *)(a))=(v))

static int sysinfo_release( struct inode *inode, struct file *file )
{
 int i;
 char *pt1,*pt2,*token,*value;
 unsigned long ultmp1;
 volatile unsigned long *regptr;
 unsigned long regval;

 pt2=pt1=sysinfo_buf;
 while ((*pt1!='=') && (*pt1!='\n') && (*pt1!='\0')) pt1++;
 if (*pt1=='=')
  { // parse option
   *pt1++='\0';
   token=pt2;
   pt2=pt1;
   while ((*pt1!='\n') && (*pt1!='\0')) pt1++;
   *pt1='\0';
   value=pt2;
   if (strcmp(token,"CMD")==0)
    {
     if (strcmp(value, "factory_off")==0)
      V_Factory_Reset = 0;
     else if (strcmp(value, "reboot")==0)
      machine_restart(NULL);
#ifdef CONFIG_BRECIS
     else if (strcmp(value, "watchdog")==0)
      {
       if (watch_dog_running)
        printk("HW watchdog is already running!\n");
       else
        {
         watch_dog_running=1;
         if (!(*WD_CTL_REG & 0x80000000)) 
          {
           printk("Turn on HW watchdog!\n");
           *WD_SVC_REG=0x5A5A5A5A;
           *WD_CTL_REG=0xFFFFFFFF;
           init_timer(&wd_timer);
           wd_timer.expires = jiffies + HZ * 10;
           wd_timer.data = 0;
           wd_timer.function = wd_refresh;
           add_timer(&wd_timer);
          }
        }
      }
     else if (strcmp(value, "nodog")==0) // test
      {
       watch_dog_running=0;
       del_timer_sync(&wd_timer);
      }
     else if (strcmp(value, "gpio")==0)
      {
       for (i=0; i<16; i++)
        {
         regptr=(volatile unsigned long *)(0xb8400180+gpio_cfg_arr[i].offset);
         regval=((*regptr)>>(gpio_cfg_arr[i].index*4))&0xf;
         printk("GPIO[%d], ",i);
         if (regval==0x00)
          printk("INPUT");
         else if (regval==0x01)
          printk("INPUT with interrupt");
         else if (regval==0x08)
          printk("OUTPUT");
         else
          printk("Special:%lx",regval);
         // GPIO value
         regptr=(volatile unsigned long *)(0xb8400170+gpio_cfg_arr[i].offset);
         regval=((*regptr>>gpio_cfg_arr[i].index)&0x1);
         printk(", VAL: %lu\n",regval);
        }
      }
     else if (strcmp(value, "mii")==0)
      {
       unsigned long base;
       unsigned long md_data,md_ca;
       int phy_reg;
       unsigned long data;
       int phy;

       for (base=0xb8600000;base<0xb8800000;base+=0x00100000)
        {
         md_data=base+0x58;
         md_ca=base+0x5c;
         printk("Check MII at base[%lx]...\n",base);
         for (phy=0;phy<6;phy++)
          {
           for (phy_reg=0;phy_reg<4;phy_reg++)
            {
             v_outl(md_ca, 0x00000800 | (phy << 5) | phy_reg);
             while (v_inl(md_ca) & 0x00000800) ;
             data = v_inl(md_data);
             printk("phy[%d], reg[%d] is %lx\n",phy,phy_reg,data);
            }
          }
        }
      }
     else if (strcmp(value, "int")==0)
      {
       printk("INT MSK: %08x\n", *INT_MSK_REG);
       printk("INT STS: %08x\n", *INT_STA_REG);
       printk("PER MSK: %08x\n", *PER_INT_MSK_REG);
       printk("PER STS: %08x\n", *PER_INT_BIT_REG);
       printk("GPIO MSK: %08x\n", *GPIO_INT_MSK_REG);
       printk("GPIO STS: %08x\n", *GPIO_INT_BIT_REG);
      }
     else if (strcmp(value, "int3o")==0)
      {
       printk("Open int3 now!\n");
       *INT_MSK_REG |= (1 << 3);
       printk("INT MSK: %08x\n", *INT_MSK_REG);
      }
     else if (strcmp(value, "int3c")==0)
      {
       printk("Close int3 now!\n");
       *INT_MSK_REG &= ~(1 << 3);
       printk("INT MSK: %08x\n", *INT_MSK_REG);
      }
     else if (strcmp(value, "extgpio")==0)
      {
       for (i=0; i<8; i++)
        printk("HC164 bit[%d]:%d\n",i,((VV_74hc164_value)&(1<<i))?1:0);
      }
#endif
    }
#ifdef CONFIG_BRECIS
   else if (strcmp(token,"wanlink")==0)
    {
     // 0: autonegotiation
     // 1: 10M Half Duplex
     // 2: 10M Full Duplex
     // 3: 100M Half Duplex
     // 4: 100M Full Duplex
     int linktmp;
     linktmp=vic_atoul(value);
     if ((linktmp<=4) && (V_wanlink!=linktmp)) // Change link setting
      {
       V_wanlink=linktmp;
       VV_SetWAN_link(V_wanlink);
      }
    }
#endif
   else if (strcmp(token,"SIP")==0) // format: SIP=contact_Port,RTP_Start,RTP_End
    {
     char *pt1;
     pt1=value;
     while ((*pt1!=',')&&(*pt1!='\0'))
      pt1++;
     if (*pt1==',')
      {
       *pt1++='\0';
       pt2=pt1;
       while ((*pt2!=',')&&(*pt2!='\0'))
        pt2++;
       if (*pt2==',')
        {
         pt2++;
         mysip_listen_port=vic_atoul(value);
         mysip_rtp_start=vic_atoul(pt1);
         mysip_rtp_end=vic_atoul(pt2);
        }
      }
    }
   else if (strcmp(token,"STUN")==0) // format: STUN=stun1,stun2... till 5
    {
     pt1=value;
     for (i=0;i<5;i++)
      {
       pt2=pt1; 
       while ((*pt2!=',')&&(*pt2!='\0'))
        pt2++;
       if (*pt2==',') 
        {
         ultmp1=1;
         *pt2++='\0';
        }
       else
        ultmp1=0;
       mysip_stun_port[i]=vic_atoul(pt1);
       if (!ultmp1)
        {
         i++; // update the record's count
         break;
        }
       pt1=pt2;
      }
     if (i<5) // not get to full records
      mysip_stun_port[i]=0; // mark the end
    }
#ifdef CONFIG_BRECIS
   else if (strcmp(token,"wifiCLled")==0)
    {
     switch (V_model)
      {
       case MODEL_AX112W_GW :
       case MODEL_AX112W :  // GPIO 6
                              regptr=(volatile unsigned long *)(0xb8400178);
                              if (value[0]=='1')
                               *regptr |= 0x01;
                              else
                               *regptr &= ~(0x01);
                              break;
       case MODEL_AX145:
       case MODEL_AX125:  // 74HC164 Q1
                              VV_74HC164_SetBit(1, (value[0]=='1')?0:1);
                              break;
      }
    }
   else if (strcmp(token,"wifiAPled")==0)
    {
     switch (V_model)
      {
       case MODEL_AX112W_GW :
       case MODEL_AX112W :  // GPIO 7
                              regptr=(volatile unsigned long *)(0xb8400178);
                              if (value[0]=='1')
                               *regptr |= 0x02;
                              else
                               *regptr &= ~(0x02);
                              break;
       case MODEL_AX145:
       case MODEL_AX125:  // 74HC164 Q0
                              VV_74HC164_SetBit(0, (value[0]=='1')?0:1);
                              break;
      }
    }
   else if (strcmp(token,"ALLled")==0) // for test program
    {
     switch (V_model)
      {
       case MODEL_AX112W_GW :
       case MODEL_AX112W :    // GPIO 5
                              regptr=(volatile unsigned long *)(0xb8400174);
                              if (value[0]=='1')
                               *regptr |= 0x08;
                              else
                               *regptr &= ~(0x08);
                              // GPIO 6-8
                              regptr=(volatile unsigned long *)(0xb8400178);
                              if (value[0]=='1')
                               *regptr |= 0x07;
                              else
                               *regptr &= ~(0x07);
                              break;
       case MODEL_AX145:
       case MODEL_AX125:    // 74HC164 Q0-Q6
                              if (value[0]=='1')
                               VV_74hc164_value &= ~(0x7f);
                              else
                               VV_74hc164_value |= 0x7f;
                              VV_74HC164_WriteOut();
                              break;
      }
    }
#endif
   else if (strcmp(token,"LANIP")==0)
    {
     pt1=value;
     while ((*pt1!=',')&&(*pt1!='\0'))
      pt1++;
     if (*pt1==',') // correct
      {
       *pt1++='\0'; 
       while (*pt1==' ') pt1++;
       V_LAN_IP=in_aton(value);
       V_LAN_MASK=in_aton(pt1);
       V_LAN_NET=V_LAN_IP & V_LAN_MASK;
//printk("VIC:set LANIP/MASK to:%u.%u.%u.%u/%u.%u.%u.%u, net seg is %u.%u.%u.%u\n", NIPQUAD(V_LAN_IP), NIPQUAD(V_LAN_MASK), NIPQUAD(V_LAN_NET));
      }
    }
#ifdef CONFIG_BRECIS
   else if (memcmp(token,"extgpio",7)==0)
    {
     ultmp1=vic_atoul(token+7);
     if (ultmp1<8) // 0-7
      VV_74HC164_SetBit(ultmp1, (value[0]=='0')?0:1);
    }
#endif
#ifdef CONFIG_IP_NF_TARGET_TRIGGER
   else if (strcmp(token,"TRIGGER_TIME")==0)
    {
     ultmp1=vic_atoul(value);
     if (ultmp1)
//{
      vic_pt_timeout=ultmp1;
//printk("VIC:set trigger timeout to :%lu\n",vic_pt_timeout);
//}
    }
#endif
  }
 sysinfo_buf[0]='\0';
 return 0;
}

static struct file_operations sysinfo_fop = {
	read:		sysinfo_read,
	write:		sysinfo_write,
        release:        sysinfo_release,
};

void sysinfo_proc_init(void)
{
 sysinfo_buf[0] = '\0';
 sysinfo_entry = create_proc_entry("sysinfo", 0, &proc_root);
 if (!sysinfo_entry)
  printk(KERN_ERR "Unable to create /proc/sysinfo\n");
 else
  sysinfo_entry->proc_fops = &sysinfo_fop;
}
////////////////////////////////////////////////////////
/////////////// End of Vic Lee, ASUS ///////////////////
////////////////////////////////////////////////////////

static int init(void * unused)
{
	struct files_struct *files;
	lock_kernel();
	do_basic_setup();

	prepare_namespace();

	/*
	 * Ok, we have completed the initial bootup, and
	 * we're essentially up and running. Get rid of the
	 * initmem segments and start the user-mode stuff..
	 */
	free_initmem();
	unlock_kernel();
	
	/*
	 * Right now we are a thread sharing with a ton of kernel
	 * stuff. We don't want to end up in user space in that state
	 */
	 
	files = current->files;
	if(unshare_files())
		panic("unshare");
	put_files_struct(files);
	
	if (open("/dev/console", O_RDWR, 0) < 0)
		printk("Warning: unable to open an initial console.\n");

	(void) dup(0);
	(void) dup(0);
	
/////////////// Add by Vic Lee, ASUS ///////////////////
#ifdef CONFIG_BRECIS
        VV_74hc164_value=0x7F;
#endif
	sysinfo_proc_init();
/////////////// End of Vic Lee, ASUS ///////////////////
	/*
	 * We try each of these until one succeeds.
	 *
	 * The Bourne shell can be used instead of init if we are 
	 * trying to recover a really broken machine.
	 */

	if (execute_command)
		run_init_process(execute_command);

	run_init_process("/sbin/init");
	run_init_process("/etc/init");
	run_init_process("/bin/init");
	run_init_process("/bin/sh");

	panic("No init found.  Try passing init= option to kernel.");
}

/////////////// Add by Vic Lee, ASUS ///////////////////
EXPORT_SYMBOL(V_model);
EXPORT_SYMBOL(VV_74HC164_SetByte);
EXPORT_SYMBOL(VV_74HC164_SetBit);
EXPORT_SYMBOL(VV_74HC164_GetBit);
EXPORT_SYMBOL(vic_atoul);
/////////////// End of Vic Lee, ASUS ///////////////////
