/* 
 * ip_conntrack_proto_ah.c - Version $Revision: 1.1 $
 *
 * Connection tracking helper module for IPSEC AH
 *
 * Based on the existing UDP helper
 *
 * (C) 2002 by Harald Welte <laforge@gnumonks.org>
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/in.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>

/* same timeouts as UDP */
#define AH_TIMEOUT (30*HZ)
#define AH_STREAM_TIMEOUT (180*HZ)

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Harald Welte <laforge@gnumonks.org>");
MODULE_DESCRIPTION("netfilter connection tracking protocol helper for AH");

static int ah_pkt_to_tuple(const void *datah, size_t datalen,
			   struct ip_conntrack_tuple *tuple)
{
	const u_int32_t *spi = datah;

	tuple->src.u.ah.spi = 0;
	tuple->dst.u.ah.spi = *spi;

	return 1;
}

static int ah_invert_tuple(struct ip_conntrack_tuple *tuple,
			   const struct ip_conntrack_tuple *orig)
{
	tuple->src.u.ah.spi = orig->dst.u.ah.spi;
	tuple->dst.u.ah.spi = orig->src.u.ah.spi;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int ah_print_tuple(char *buffer,
				   const struct ip_conntrack_tuple *tuple)
{
	return sprintf(buffer, "spi=%u ", ntohs(tuple->dst.u.ah.spi));
}

/* Print out the private part of the conntrack. */
static unsigned int ah_print_conntrack(char *buffer,
				       const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static int ah_packet(struct ip_conntrack *conntrack,
		     struct iphdr *iph, size_t len,
		     enum ip_conntrack_info conntrackinfo)
{
	/* If we've seen traffic both ways, this is some kind of AH
	   stream.  Extend timeout. */
	if (conntrack->status & IPS_SEEN_REPLY) {
		ip_ct_refresh(conntrack, AH_STREAM_TIMEOUT);
		/* Also, more likely to be important, and not a probe */
		set_bit(IPS_ASSURED_BIT, &conntrack->status);
	} else
		ip_ct_refresh(conntrack, AH_TIMEOUT);

	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int ah_new(struct ip_conntrack *conntrack,
		  struct iphdr *iph, size_t len)
{
	return 1;
}

struct ip_conntrack_protocol ip_ct_p_ah
= { { NULL, NULL }, IPPROTO_AH, "ah",
    ah_pkt_to_tuple, ah_invert_tuple, ah_print_tuple, ah_print_conntrack,
    ah_packet, ah_new, NULL, NULL, THIS_MODULE };


static int __init init(void)
{
	int ret;
	if ((ret = ip_conntrack_protocol_register(&ip_ct_p_ah))) {
		printk(KERN_ERR "unable to register conntrack protocol "
				"helper for AH: %d\n", ret);
		return -EIO;
	}
	return 0;
}

static void __exit fini(void)
{
	ip_conntrack_protocol_unregister(&ip_ct_p_ah);
}

module_init(init);
module_exit(fini);

