/* 
 * ip_conntrack_proto_esp.c - Version $Revision: 1.1 $
 *
 * Connection tracking helper module for IPSEC ESP
 *
 * Based on the existing UDP helper
 *
 * (C) 2002 by Harald Welte <laforge@gnumonks.org>
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/in.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>

/* same timeouts as UDP */
#define ESP_TIMEOUT (30*HZ)
#define ESP_STREAM_TIMEOUT (180*HZ)

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Harald Welte <laforge@gnumonks.org>");
MODULE_DESCRIPTION("netfilter connection tracking protocol helper for ESP");

static int esp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_conntrack_tuple *tuple)
{
	const u_int32_t *spi = datah;

	// Fixed by Vic Lee, ASUS
	//tuple->src.u.esp.spi = 0;
	tuple->src.u.esp.spi = *spi;
	tuple->dst.u.esp.spi = *spi;

	return 1;
}

static int esp_invert_tuple(struct ip_conntrack_tuple *tuple,
			    const struct ip_conntrack_tuple *orig)
{
	tuple->src.u.esp.spi = orig->dst.u.esp.spi;
	tuple->dst.u.esp.spi = orig->src.u.esp.spi;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int esp_print_tuple(char *buffer,
				    const struct ip_conntrack_tuple *tuple)
{
	return sprintf(buffer, "spi=%u ", ntohs(tuple->src.u.esp.spi));
}

/* Print out the private part of the conntrack. */
static unsigned int esp_print_conntrack(char *buffer,
					const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static int esp_packet(struct ip_conntrack *conntrack,
		      struct iphdr *iph, size_t len,
		      enum ip_conntrack_info conntrackinfo)
{
	/* If we've seen traffic both ways, this is some kind of ESP
	   stream.  Extend timeout. */
	if (conntrack->status & IPS_SEEN_REPLY) {
		ip_ct_refresh(conntrack, ESP_STREAM_TIMEOUT);
		/* Also, more likely to be important, and not a probe */
		set_bit(IPS_ASSURED_BIT, &conntrack->status);
	} else
		ip_ct_refresh(conntrack, ESP_TIMEOUT);

	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int esp_new(struct ip_conntrack *conntrack,
			     struct iphdr *iph, size_t len)
{
	return 1;
}

struct ip_conntrack_protocol ip_ct_p_esp
= { { NULL, NULL }, IPPROTO_ESP, "esp",
    esp_pkt_to_tuple, esp_invert_tuple, esp_print_tuple, esp_print_conntrack,
    esp_packet, esp_new, NULL, NULL, THIS_MODULE };

static int __init init(void)
{
	int ret;

	if ((ret = ip_conntrack_protocol_register(&ip_ct_p_esp))) {
		printk(KERN_ERR "unable to register conntrack protocol "
				"helper for ESP: %d\n", ret);
		return -EIO;
	}
	return 0;
}

static void __exit fini(void)
{
	ip_conntrack_protocol_unregister(&ip_ct_p_esp);
}

module_init(init);
module_exit(fini);
