#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_mypass.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

static int
ipt_mypass_match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	struct ipt_mypassinfo *info = (struct ipt_mypassinfo *)matchinfo;
	struct ip_conntrack *ct;
	enum ip_conntrack_info ctinfo;

	ct=ip_conntrack_get((struct sk_buff *)skb, &ctinfo);
	if (ct) {
		if (info->act_type==MYPASS_CHK) {
			DEBUGP("Vic MYPASS check ct[%p], its value[%d], ", ct, ct->v_pass_mark);
			if ((ct->v_pass_mark & info->value)==info->value) {
				DEBUGP("PASS!!\n");
				return 1;
			} else {
				DEBUGP("DROP!!\n");
				return 0;
			}
		}
		else if (info->act_type==MYPASS_SET) {
			DEBUGP("Vic MYPASS ct[%p] set value[%d]!!\n", ct, info->value);
			ct->v_pass_mark=info->value;
			return 0; // always not match
		}
		else if (info->act_type==MYPASS_BOR) {
			DEBUGP("Vic MYPASS ct[%p] before OR value [%d]!!\n", ct, ct->v_pass_mark);
			ct->v_pass_mark|=info->value;
			DEBUGP("Vic MYPASS ct[%p] after OR[%d], become [%d]!!\n", ct, info->value, ct->v_pass_mark);
		} else { // should be MYPASS_BCL
			DEBUGP("Vic MYPASS ct[%p] before CLEAR value [%d]!!\n", ct, ct->v_pass_mark);
			ct->v_pass_mark&=(~info->value);
			DEBUGP("Vic MYPASS ct[%p] after CLEAR[%d], become [%d]!!\n", ct, info->value, ct->v_pass_mark);
		}
	}
	return 0;
}

/* Called when user tries to insert an entry of this type. */
static int
ipt_mypass_checkentry(const char *tablename,
	   const struct ipt_ip *ip,
	   void *matchinfo,
	   unsigned int matchsize,
	   unsigned int hook_mask)
{
	struct ipt_mypassinfo *info = (struct ipt_mypassinfo *)matchinfo;
	if (matchsize != IPT_ALIGN(sizeof(struct ipt_mypassinfo)))
		return 0;

	if ((info->act_type!=MYPASS_CHK)&&(info->act_type!=MYPASS_SET)&&(info->act_type!=MYPASS_BOR)&&(info->act_type!=MYPASS_BCL))
		return 0;

	return 1;
}

static struct ipt_match ipt_mypass_reg
= { { NULL, NULL }, "mypass", &ipt_mypass_match, &ipt_mypass_checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	return ipt_register_match(&ipt_mypass_reg);
}

static void __exit fini(void)
{
	ipt_unregister_match(&ipt_mypass_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
