#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/netfilter_ipv4/ip_conntrack.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_mysip.h>

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

extern unsigned short mysip_listen_port;
extern unsigned short mysip_rtp_start, mysip_rtp_end;
extern unsigned short mysip_stun_port[5];

static int
ipt_mysip_match(const struct sk_buff *skb,
      const struct net_device *in,
      const struct net_device *out,
      const void *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)matchinfo;
	struct iphdr *iph;
	struct udphdr *udph;
	unsigned short port;

	iph = skb->nh.iph;
	if (iph->protocol != IPPROTO_UDP) // we just use udp till now
		return 0;
	if (datalen < sizeof (*udph)) // incomplete packet
		return 0;
	udph = (struct udphdr *)((char *)iph + (iph->ihl<<2));
	if (sinfo->chk_type==MYSIP_CHK_DST) {
		port=ntohs(udph->dest);
	} else { /* should be MYSIP_CHK_SRC */
		port=ntohs(udph->source);
	}
        DEBUGP("+++MYSIP+++  src:%u, dst:%u !!!!!!!!!!!!!!, check port is:%u, type is %s\n",udph->source,udph->dest,port,(info->chk_type==MYSIP_CHK_DST)?"Dst":"Src");
	// Check if it is our intresting port
	if (port!=mysip_listen_port) {
		if ((port<mysip_rtp_start)||(port>mysip_rtp_end)) {
			int i, matched=0;
			for (i=0;(i<5)&&(mysip_stun_port[i]!=0);i++) {
				if (port==mysip_stun_port[i]) {
					matched=1;
					break;
				}
			}
			if (!matched)
				return 0;
		}
	}
	// Got our intresting port, set the mark if needed
	if (sinfo->is_set_mark) {
		struct ip_conntrack *ct;
		enum ip_conntrack_info ctinfo;
		ct=ip_conntrack_get((struct sk_buff *)skb, &ctinfo);
		if (ct) {
			DEBUGP("Enter Vic MYSIP set mark, ct :%p!!!!!!!!!!!!!!\n",ct);
			ct->v_qos_mark=DEFAULT_MYSIP_MARK;
		}
	}
	DEBUGP("Vic MYSIP Accept !!!!!!!!!!!!!!\n");
	return 1;
}

/* Called when user tries to insert an entry of this type. */
static int
ipt_mysip_checkentry(const char *tablename,
	   const struct ipt_ip *ip,
	   void *matchinfo,
	   unsigned int matchsize,
	   unsigned int hook_mask)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)matchinfo;
	if (matchsize != IPT_ALIGN(sizeof(struct ipt_mysipinfo)))
		return 0;

	if ((sinfo->chk_type!=MYSIP_CHK_DST)&&(sinfo->chk_type!=MYSIP_CHK_SRC))
		return 0;

	return 1;
}

static struct ipt_match ipt_mysip_reg
= { { NULL, NULL }, "mysip", &ipt_mysip_match, &ipt_mysip_checkentry, NULL, THIS_MODULE };

static int __init init(void)
{
	return ipt_register_match(&ipt_mysip_reg);
}

static void __exit fini(void)
{
	ipt_unregister_match(&ipt_mysip_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
