/* Shared library add-on to iptables for mypass. */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_mypass.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"mypass v%s options:\n"
"--act_type set|chk|bor|bcl\n"
"--val val[0-255]\n"
"\n", IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "act_type", 1, 0, 'c' },
	{ "val", 1, 0, 's' },
	{ 0 }
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	struct ipt_mypassinfo *sinfo = (struct ipt_mypassinfo *)m->data;

	sinfo->act_type = MYPASS_CHK;
	sinfo->value = 1; // initial value

	/* Can't cache this. */
	*nfcache |= NFC_UNKNOWN;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_mypassinfo *sinfo = (struct ipt_mypassinfo *)(*match)->data;
	int val;

	switch(c) {
	case 'c':
		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --act_type");
		if (strcmp(optarg,"chk")==0)
			sinfo->act_type=MYPASS_CHK;
		else if (strcmp(optarg,"set")==0)
			sinfo->act_type=MYPASS_SET;
		else if (strcmp(optarg,"bor")==0)
			sinfo->act_type=MYPASS_BOR;
		else if (strcmp(optarg,"bcl")==0)
			sinfo->act_type=MYPASS_BCL;
		else
			exit_error(PARAMETER_PROBLEM, "Invalid act_type");
		break;

	case 's':
		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --val");
		val=atoi(optarg);
		if ((val>255)||(val<0))
			exit_error(PARAMETER_PROBLEM, "Invalid value");
		else
			sinfo->value=(u_int8_t)val;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; must have specified --mac. */
static void final_check(unsigned int flags)
{
}

static char *act_type_name(u_int8_t type)
{
	char *ppt;
	switch (type) {
	default:
	case MYPASS_CHK : ppt="chk";
			  break;
	case MYPASS_SET : ppt="set";
			  break;
	case MYPASS_BOR : ppt="bor";
			  break;
	case MYPASS_BCL : ppt="bcl";
			  break;
	}
	return ppt;
}
/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	struct ipt_mypassinfo *sinfo = (struct ipt_mypassinfo *)match->data;
	printf("act_type:%s , val:%u", act_type_name(sinfo->act_type), sinfo->value);
}

/* FIXME: Make minimalist: only print rate if not default --RR */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	struct ipt_mypassinfo *sinfo = (struct ipt_mypassinfo *)match->data;
	printf("--act_type %s --val %u", act_type_name(sinfo->act_type), sinfo->value);
}


static
struct iptables_match mypass
= { NULL,
    "mypass",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_mypassinfo)),
    IPT_ALIGN(sizeof(struct ipt_mypassinfo)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&mypass);
}
