/* Shared library add-on to iptables for mysip. */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ipt_mysip.h>

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"mysip v%s options:\n"
"--chk_type dst|src\n"
"--setmark\n"
"\n", IPTABLES_VERSION);
}

static struct option opts[] = {
	{ "chk_type", 1, 0, 'c' },
	{ "setmark", 0, 0, 's' },
	{ 0 }
};

/* Initialize the match. */
static void
init(struct ipt_entry_match *m, unsigned int *nfcache)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)m->data;

	sinfo->chk_type = MYSIP_CHK_DST;
	sinfo->is_set_mark = 0; // initial value

	/* Can't cache this. */
	*nfcache |= NFC_UNKNOWN;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      unsigned int *nfcache,
      struct ipt_entry_match **match)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)(*match)->data;

	switch(c) {
	case 'c':
		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --chk_type");
		if (strcmp(optarg,"dst")==0)
			sinfo->chk_type=MYSIP_CHK_DST;
		else if (strcmp(optarg,"src")==0)
			sinfo->chk_type=MYSIP_CHK_SRC;
		else
			exit_error(PARAMETER_PROBLEM, "Invalid chk_type");
		break;

	case 's':
		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --setmark");
		sinfo->is_set_mark = 1;
		break;

	default:
		return 0;
	}

	return 1;
}

/* Final check; must have specified --mac. */
static void final_check(unsigned int flags)
{
}

/* Prints out the matchinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_match *match,
      int numeric)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)match->data;
	printf("chk_type: %s , ", (sinfo->chk_type==MYSIP_CHK_DST)?"dst":"src");
	printf("setmark %s ", (sinfo->is_set_mark)?"yes":"no");
}

/* FIXME: Make minimalist: only print rate if not default --RR */
static void save(const struct ipt_ip *ip, const struct ipt_entry_match *match)
{
	struct ipt_mysipinfo *sinfo = (struct ipt_mysipinfo *)match->data;

	printf("--chk_type %s ", (sinfo->chk_type==MYSIP_CHK_DST)?"dst":"src");
	if (sinfo->is_set_mark)
		printf("--setmark ");
}


static
struct iptables_match mysip
= { NULL,
    "mysip",
    IPTABLES_VERSION,
    IPT_ALIGN(sizeof(struct ipt_mysipinfo)),
    IPT_ALIGN(sizeof(struct ipt_mysipinfo)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
	register_match(&mysip);
}
