#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <string.h>

#define PPP_BINARY	"pppd"
#define PPTP_RETRY_INTERVAL	10	/* second */

#define DEFAULT_ETH	"eth1"

#ifndef IFNAMSIZ
#define IFNAMSIZ	16
#endif

#ifndef _PATH_VARRUN
#define _PATH_VARRUN 	"/var/run/"
#endif


char mlinkdevname[IFNAMSIZ];
int  mlinkdevorder = 0;

pid_t pid;
int do_exit = 0;

void launch_prog(int argc, char **argv) {
  char *new_argv[24]; /* XXX if not using GCC, hard code a limit here. */
  int i = 0, j;

  for (j=0; j<argc; j++)
    new_argv[i++] = argv[j];
  
  new_argv[i++] = "nodetach";	/* pppd should use this to make pppd exit can be trigger by pppoecd process */
  new_argv[i] = NULL;

  new_argv[0] = PPP_BINARY;
  
  execvp(new_argv[0], new_argv); /* excute a file */

}

int parse_username(int argc, char **argv) {
	int j;

	for (j=0; j<argc-1; j++)
	{
		if( argv[j] && argv[j+1] 
			&& (strcmp(argv[j], "user") == 0) 
			&& (strlen(argv[j+1]) > 0 )
			)
		{
			return 1;
		}
	}
	return 0;
}
void parse_devname(int argc, char **argv) {

	int j;

	strcpy(mlinkdevname, DEFAULT_ETH);

	for (j=0; j<argc-1; j++)
	{
		if( argv[j] && argv[j+1] && (strncmp(argv[j], "mlinkdevname",12) == 0) &&
			(strncmp(argv[j+1], "eth",3)==0) )
		{
			strncpy(mlinkdevname,argv[j+1],sizeof(mlinkdevname)-1);
		}
		if( argv[j] && argv[j+1] && (strncmp(argv[j], "mlinkdevorder",12) == 0) )
		{
			mlinkdevorder = atoi(argv[j+1]);
		}

	}
}

void donothing(int i)
{
}
void sighandler(int signo)
{	
	do_exit =1;
	kill(pid,SIGTERM);
	sleep(1);
	//kill(pid,SIGKILL);
	exit(0);
}

/*
 * Create a file containing our process ID.
 */
static void
create_pidfile(pid)
    int pid;
{
    FILE *pidfile;
    char pidfilename[64];
	
    snprintf(pidfilename, sizeof(pidfilename), "%s/pppoe-%s.%d.pid",
	     _PATH_VARRUN,  mlinkdevname, mlinkdevorder );
    if ((pidfile = fopen(pidfilename, "w")) != NULL) {
	fprintf(pidfile, "%d\n", pid);
	(void) fclose(pidfile);
    } else {
	printf("Failed to create pid file %s\n", pidfilename);
	pidfilename[0] = 0;
    }
}

int
main(int argc, char **argv)
{
	int status;

  	memset(mlinkdevname,0,sizeof(mlinkdevname));
  	daemon(0,0); 
	//make it to be new process group to avoid parent process be killed by kill(0,SIGTERM) from child process
	
	signal(SIGINT,  sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGKILL, sighandler);
	signal(SIGPIPE, sighandler);
	signal(SIGHUP,  sighandler);
	signal(SIGUSR1, donothing); /* avoid child kill me */
	signal(SIGUSR2, donothing); /* avoid child kill me */
	//signal(SIGCHLD, SIG_IGN); we want to wait for it

	if( parse_username(argc, argv ) == 0 )
		return -1;

	parse_devname(argc, argv);
	create_pidfile(getpid());

	if( argc > 1 && strcmp(argv[1],"0.0.0.0")==0 )
		return -1;
	
	for(;; )
	{
		switch( (pid=fork()) )
		{
		case -1:
			exit(-1);
		case 0: /* child */
			launch_prog(argc, argv);
			printf("pppoecd fail to run pppd\n");fflush(stdout);
			exit(-1);
		default: /* parent */
			while( wait( &status ) != pid )
				;
			if( do_exit )
			{
				kill(pid,SIGTERM);
				kill(0,SIGTERM);
				exit(0);
			}
			sleep(PPTP_RETRY_INTERVAL);
			break;
		}
	}
	
	return 0;
}
