#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <string.h>

#define PPP_BINARY	"pptp"
#define PPTP_RETRY_INTERVAL	10	/* second */

#define DEFAULT_ETH	"eth1"

#ifndef IFNAMSIZ
#define IFNAMSIZ	16
#endif

#ifndef _PATH_VARRUN
#define _PATH_VARRUN 	"/var/run/"
#endif


char mlinkdevname[IFNAMSIZ];
int  mlinkdevorder = 0;

pid_t pid;
int do_exit = 0;

void launch_prog(int argc, char argv[24][32]) 
{
  char *new_argv[24]; /* XXX if not using GCC, hard code a limit here. */
  int i = 0, j;

  for (j=0; j<argc; j++)
    new_argv[i++] = argv[j];

#if 1
  new_argv[i++] = "ppp_mode";	/* let pppd know that it is used for pptp, so, don't load pppoe module and don't loop forever */
  new_argv[i++] = "1";
  new_argv[i++] = "nodetach";	/* pptp should use this to make pppd exit can be trigger by pptp process */
#endif
  
  new_argv[i++] = "nodetach";	/* pptp should use this to make pppd exit can be trigger by pptp process */
  new_argv[i] = NULL;

  new_argv[0] = PPP_BINARY;
  
  execvp(new_argv[0], new_argv); /* excute a file */

}

int parse_username(int argc, char argv[24][32]) 
{
	int j;

	for (j=0; j<argc-1; j++)
	{
		if( argv[j] && argv[j+1] 
			&& (strcmp(argv[j], "user") == 0) 
			&& (strlen(argv[j+1]) > 0 )
			)
		{
			return 1;
		}
	}
	return 0;
}
void parse_devname(int argc, char argv[24][32]) 
{

	int i;
	for (i = 0 ; i < argc - 1 ; i++) {
		if ( argv[i] && argv[i+1] &&
			(strncmp (argv[i], "linkname", 8) == 0)) {
			snprintf(mlinkdevname, sizeof(mlinkdevname), "eth0.%d", atoi(argv[i+1]) + 2);
			break;
		}
	}

#if 0
	int j;

	strcpy(mlinkdevname, DEFAULT_ETH);

	for (j=0; j<argc-1; j++)
	{
		if( argv[j] && argv[j+1] && (strncmp(argv[j], "mlinkdevname",12) == 0) &&
			(strncmp(argv[j+1], "eth",3)==0) )
		{
			strncpy(mlinkdevname,argv[j+1],sizeof(mlinkdevname)-1);
			break;
		}
	}
#endif
}

void donothing(int i)
{
}
void sighandler(int signo)
{	
	do_exit =1;
	kill(pid,SIGTERM);
	//sleep(1);
	//kill(pid,SIGKILL);
	exit(0);
}

/*
 * Create a file containing our process ID.
 */
static void
create_pidfile(pid)
    int pid;
{
    FILE *pidfile;
    char pidfilename[64];
	
    snprintf(pidfilename, sizeof(pidfilename), "%s/pptp-%s.%d.pid",
	     _PATH_VARRUN,  mlinkdevname, mlinkdevorder );
    if ((pidfile = fopen(pidfilename, "w")) != NULL) {
	fprintf(pidfile, "%d\n", pid);
	(void) fclose(pidfile);
    } else {
	printf("Failed to create pid file %s\n", pidfilename);
	pidfilename[0] = 0;
    }
}

int main(int argc, char **argv)
{
	int status;
	int loc_argc = argc;
	char loc_argv[24][32];	/* XXX if not using GCC, hard code a limit here. */
	int i = 0, j;

	memset(loc_argv,0,sizeof(loc_argv));
	for (j=0; j<argc; j++)
		strncpy(loc_argv[i++] , argv[j] , sizeof(loc_argv[0]));
	
  	memset(mlinkdevname,0,sizeof(mlinkdevname));
  	daemon(0,0); 
	//make it to be new process group to avoid parent process be killed by kill(0,SIGTERM) from child process
	
	signal(SIGINT,  sighandler);
	signal(SIGTERM, sighandler);
	signal(SIGKILL, sighandler);
	signal(SIGPIPE, sighandler);
	signal(SIGHUP,  sighandler);
	signal(SIGUSR1, donothing); /* avoid child kill me */
	signal(SIGUSR2, donothing); /* avoid child kill me */

	//signal(SIGCHLD, SIG_IGN); we want to wait for it

	if( parse_username(loc_argc, loc_argv ) == 0 )
		return -1;

	parse_devname(loc_argc, loc_argv);
	create_pidfile(getpid());

	if( loc_argc > 1 && strcmp(loc_argv[1],"0.0.0.0")==0 )
		return -1;
	
	for(;; )
	{
		switch( (pid=fork()) )
		{
		case -1:
			exit(-1);
		case 0: /* child */
			launch_prog(loc_argc, loc_argv);
			printf("pptpmain fail to run pptp\n");fflush(stdout);
			exit(-1);
		default: /* parent */
			while( wait( &status ) != pid )
				;
			if( do_exit )
			{
				kill(pid,SIGTERM);
				kill(0,SIGTERM);
				exit(0);
			}
			break;
		}
		sleep(PPTP_RETRY_INTERVAL);
	}
	
	return 0;
}
