///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2003 C++ to C Made by PC32 Lsshi
// All rights reserved. 

///////////////////////////////////////////////////////////////////////////

#include <string.h>
#include <syslog.h>
#include <pthread.h>
#include "gate.h"
#include "main.h"
#include "sample_util.h"
#include "portmap.h"
#include <arpa/inet.h>

#include <unistd.h>

#define UpnpDocument_free ixmlDocument_free
#define UpnpParse_Buffer ixmlParseBuffer
typedef IXML_Document *Upnp_Document;

pthread_mutex_t DevMutex = PTHREAD_MUTEX_INITIALIZER;

char *GateDeviceType[] = {"urn:schemas-upnp-org:device:InternetGatewayDevice:1"
                        ,"urn:schemas-upnp-org:device:WANDevice:1"
                         ,"urn:schemas-upnp-org:device:WANConnectionDevice:1"};

char *GateServiceType[] = {"urn:schemas-microsoft-com:service:OSInfo:1"
                          ,"urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1"
                          ,"urn:schemas-upnp-org:service:WANIPConnection:1"};

char *GateServiceId[] = {"urn:microsoft-com:serviceId:OSInfo1"
                        ,"urn:upnp-org:serviceId:WANCommonIFC1"
                        ,"urn:upnp-org:serviceId:WANIPConn1"};

int getProtoNum(char * proto)
{
	if (strcmp(proto,"TCP")==0)
                return 6;
        else if (strcmp(proto,"UDP")==0)
                return 17;
	else return 0;
}
char *getProtoName(int prt)
{
	switch (prt)
	{
		case 6:
			return "TCP";
			break;
		case 17:
			return "UDP";
			break;
		default:
			return "TCP";
	}
}

int chkIPADDRstring(char * addr) {
        struct in_addr dmy;
        int ip;
        
        ip=inet_aton(addr, &dmy);
        
        //printf("ip=%d\n",ip);
        
        return ip;
}

void _Gate_Void_Init(struct _Gate* pGate)
{
}


void _Gate_Void_Destroy(struct _Gate* pGate)
{
	if (pGate->gate_udn)
	{
		free(pGate->gate_udn);
		pGate->gate_udn=NULL;
	}
		
	if (pGate->m_ipcon)
	{
		pGate->m_ipcon->IPCon_Void_Destroy(pGate->m_ipcon);
		free(pGate->m_ipcon);
		pGate->m_ipcon=NULL;
	}
}

int _GateDeviceCallbackEventHandler(struct _Gate* pGate,Upnp_EventType EventType, void *Event, void *Cookie)
{
	switch ( EventType)
	{

		case UPNP_EVENT_SUBSCRIPTION_REQUEST:
		{
			_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
			pGate->GateDeviceHandleSubscriptionRequest(pGate,(struct Upnp_Subscription_Request *) Event);
			break;
		}

		case UPNP_CONTROL_GET_VAR_REQUEST:
		{
			_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
			pGate->GateDeviceHandleGetVarRequest((struct Upnp_State_Var_Request *) Event);
			break;
		}

		case UPNP_CONTROL_ACTION_REQUEST:
		{
			_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
			pGate->GateDeviceHandleActionRequest(pGate,(struct Upnp_Action_Request *) Event);
			break;
		}

		default:

			SYS_DEBUG(LOG_DEBUG, "Error in DeviceCallbackEventHandler: unknown event type %d\n", EventType);
	}

	return(0);
}

int _GateDeviceStateTableInit(struct _Gate* pGate,char *DescDocURL)
{
	Upnp_Document DescDoc = NULL;
	int ret = UPNP_E_SUCCESS;
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));

	if (UpnpDownloadXmlDoc(DescDocURL, &DescDoc) != UPNP_E_SUCCESS)
	{
		SYS_DEBUG(LOG_ERR, "DeviceStateTableInit -- Error Parsing %s\n", DescDocURL);
		ret =UPNP_E_INVALID_DESC;
	}

	pGate->gate_udn = SampleUtil_GetFirstDocumentItem(DescDoc, "UDN");

	return (ret);
}

int _GateDeviceHandleSubscriptionRequest(struct _Gate* pGate,struct Upnp_Subscription_Request *sr_event)
{
	Upnp_Document PropSet;
	char *address = NULL;
	
	PropSet = NULL;
	
	address = pGate->m_ipcon->IPCon_GetIpAddrStr(pGate->m_ipcon);
	
	//_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));

	pthread_mutex_lock(&DevMutex);

	if (strcmp(sr_event->UDN, pGate->gate_udn) == 0)
	{
		if (strcmp(sr_event->ServiceId, GateServiceId[GATE_SERVICE_OSINFO]) == 0)
		{
			UpnpAddToPropertySet(&PropSet, "OSMajorVersion","5");
			UpnpAddToPropertySet(&PropSet, "OSMinorVersion","1");
			UpnpAddToPropertySet(&PropSet, "OSBuildNumber","2600");
			UpnpAddToPropertySet(&PropSet, "OSMachineName",MACHINE_NAME);
			UpnpAcceptSubscriptionExt(pGate->device_handle, sr_event->UDN,
					sr_event->ServiceId, PropSet, sr_event->Sid);
			UpnpDocument_free(PropSet);


		}
		else if (strcmp(sr_event->ServiceId, GateServiceId[GATE_SERVICE_CONFIG]) == 0)
		{
			UpnpAddToPropertySet(&PropSet, "PhysicalLinkStatus", pGate->m_ipcon->IPCon_GetIfStatStr(pGate->m_ipcon));
			UpnpAcceptSubscriptionExt(pGate->device_handle, sr_event->UDN, sr_event->ServiceId, PropSet, sr_event->Sid);
			UpnpDocument_free(PropSet);
		}
		else if (strcmp(sr_event->ServiceId, GateServiceId[GATE_SERVICE_CONNECT])==0)
		{
			UpnpAddToPropertySet(&PropSet, "PossibleConnectionTypes","IP_Routed");
			UpnpAddToPropertySet(&PropSet, "ConnectionStatus","Connected");
			UpnpAddToPropertySet(&PropSet, "X_Name", X_NAME);
			UpnpAddToPropertySet(&PropSet, "ExternalIPAddress",address);
			UpnpAddToPropertySet(&PropSet, "PortMappingNumberOfEntries","0");
			UpnpAcceptSubscriptionExt(pGate->device_handle, sr_event->UDN,	sr_event->ServiceId, PropSet, sr_event->Sid);
			UpnpDocument_free(PropSet);
		}
	}
	pthread_mutex_unlock(&DevMutex);
	if (address) free(address);
	return(1);
}

int _GateDeviceHandleGetVarRequest(struct Upnp_State_Var_Request *cgv_event)
{
	int getvar_succeeded = 0;
	cgv_event->CurrentVal = NULL;
	
	pthread_mutex_lock(&DevMutex);

#if 1	
	if (strcmp(cgv_event->StateVarName,"ConnectionStatus") == 0)//lsshi add 2003-8-22 14:06
	{
		
	        cgv_event->CurrentVal =(char*)malloc(20);
	        
	        if(cgv_event->CurrentVal)
	        {
	        	strcpy(cgv_event->CurrentVal,"Connected");
	        	cgv_event->ErrCode = UPNP_E_SUCCESS;
	        	pthread_mutex_unlock(&DevMutex);
	        	return(cgv_event->ErrCode);
	        }
	        
	    
		cgv_event->ErrCode = 404;
		strcpy(cgv_event->ErrStr, "Invalid Variable");
		pthread_mutex_unlock(&DevMutex);
	       return(cgv_event->ErrCode);
		
		
	}
#endif

        if (getvar_succeeded)
       	{
	        cgv_event->ErrCode = UPNP_E_SUCCESS;
		}
       	else
       	{
		SYS_DEBUG(LOG_ERR,"Error in UPNP_CONTROL_GET_VAR_REQUEST callback:\n");
		SYS_DEBUG(LOG_ERR,"   Unknown variable name = %s\n",  cgv_event->StateVarName);
		cgv_event->ErrCode = 404;
		strcpy(cgv_event->ErrStr, "Invalid Variable");
        }
	pthread_mutex_unlock(&DevMutex);
	
	return(cgv_event->ErrCode == UPNP_E_SUCCESS);
}

int _GateDeviceHandleActionRequest(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	int result = 0;
	
	pthread_mutex_lock(&DevMutex);
	
	//printf("\nserviceid=%s ", ca_event->ServiceID );
	//printf("\nactionn  =%s ", ca_event->ActionName );
	
	if( strcmp(ca_event->ActionName,"GetExternalIPAddress") == 0)
	{
		result = pGate->GateDeviceGetExternalIPAddress(pGate,ca_event);
		//printf("\nresult = %d \n", result );
	}
	
	if (strcmp(ca_event->DevUDN, pGate->gate_udn) == 0)
	{
		if (strcmp(ca_event->ServiceID,"urn:upnp-org:serviceId:WANIPConn1") ==0)
		{
				_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
				
				if (strcmp(ca_event->ActionName,"SetConnectionType") == 0)
				     	result = pGate->GateDeviceSetConnectionType(ca_event);
				else if (strcmp(ca_event->ActionName,"GetConnectionTypeInfo") == 0)
					result = pGate->GateDeviceGetConnectionTypeInfo(ca_event);
				else if (strcmp(ca_event->ActionName,"RequestConnection") == 0)
					result = pGate->GateDeviceRequestConnection(ca_event);
				else if (strcmp(ca_event->ActionName,"RequestTermination") == 0)
					result = pGate->GateDeviceRequestTermination(ca_event);
				else if (strcmp(ca_event->ActionName,"ForceTermination") == 0)
					result = pGate->GateDeviceForceTermination(ca_event);
				else if (strcmp(ca_event->ActionName,"SetAutoDisconnectTime") == 0)
					result = pGate->GateDeviceSetAutoDisconnectTime(ca_event);
				else if (strcmp(ca_event->ActionName,"SetIdleDisconnectTime") == 0)
					result = pGate->GateDeviceSetIdleDisconnectTime(ca_event);
				else if (strcmp(ca_event->ActionName,"SetWarnDisconnectDelay") == 0)
					result = pGate->GateDeviceSetWarnDisconnectDelay(ca_event);
				else if (strcmp(ca_event->ActionName,"GetStatusInfo") == 0)
					result = pGate->GateDeviceGetStatusInfo(pGate,ca_event);
				else if (strcmp(ca_event->ActionName,"GetAutoDisconnectTime") == 0)
					result = pGate->GateDeviceGetAutoDisconnectTime(ca_event);
				else if (strcmp(ca_event->ActionName,"GetIdleDisconnectTime") == 0)
					result = pGate->GateDeviceGetIdleDisconnectTime(ca_event);
				else if (strcmp(ca_event->ActionName,"GetWarnDisconnectDelay") == 0)
					result = pGate->GateDeviceGetWarnDisconnectDelay(ca_event);
				else if (strcmp(ca_event->ActionName,"GetNATRSIPStatus") == 0)
					result = pGate->GateDeviceGetNATRSIPStatus(ca_event);
				else if (strcmp(ca_event->ActionName,"GetGenericPortMappingEntry") == 0)
					result = pGate->GateDeviceGetGenericPortMappingEntry(ca_event);
				else if (strcmp(ca_event->ActionName,"GetSpecificPortMappingEntry") == 0)
					result = pGate->GateDeviceGetSpecificPortMappingEntry(ca_event);
				else if (strcmp(ca_event->ActionName,"GetExternalIPAddress") == 0)
					result = pGate->GateDeviceGetExternalIPAddress(pGate,ca_event);
				else if (strcmp(ca_event->ActionName,"AddPortMapping") == 0)
					result = pGate->GateDeviceAddPortMapping(pGate,ca_event);
				else if (strcmp(ca_event->ActionName,"DeletePortMapping") == 0)
					result = pGate->GateDeviceDeletePortMapping(pGate,ca_event);
				else result = 0;
		}
		else if (strcmp(ca_event->ServiceID,GateServiceId[GATE_SERVICE_CONFIG]) == 0)
		{
#if 1//lsshi 8/26/2003 10:57AM
			if (strcmp(ca_event->ActionName,"GetCommonLinkProperties") == 0)
				result = pGate->GateDeviceGetCommonLinkProperties(ca_event);
			else if (strcmp(ca_event->ActionName, "X_GetICSStatistics") == 0)	//lsshi 2003-8-15 12:05 "X"
				result = pGate->GateDeviceX(ca_event);
			else if (strcmp(ca_event->ActionName,"GetTotalBytesSent") == 0)
				result = pGate->GateDeviceGetTotalBytesSent(pGate,ca_event);
			else if (strcmp(ca_event->ActionName,"GetTotalBytesReceived") == 0)
				result = pGate->GateDeviceGetTotalBytesReceived(pGate,ca_event);
			else if (strcmp(ca_event->ActionName,"GetTotalPacketsSent") == 0)
				result = pGate->GateDeviceGetTotalPacketsSent(pGate,ca_event);
			else if (strcmp(ca_event->ActionName,"GetTotalPacketsReceived") == 0)
				result = pGate->GateDeviceGetTotalPacketsReceived(pGate,ca_event);
			else result = 0;
#else
			ca_event->ErrCode = 401;
			strcpy(ca_event->ErrStr, "Invalid Action");
			ca_event->ActionResult = NULL;
			
			//_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
			result=ca_event->ErrCode;
			
#endif
		}
	}
	else
		printf("\nDevUDN != gate_udn (%s != %s) ", ca_event->DevUDN , pGate->gate_udn);
	
	
	pthread_mutex_unlock(&DevMutex);

	return(result);
}

int _GateDeviceX(struct Upnp_Action_Request *ca_event)
{
	ca_event->ErrCode = 401;
	strcpy(ca_event->ErrStr, "Invalid Action");
	ca_event->ActionResult = NULL;
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return (ca_event->ErrCode);
}

int _GateDeviceGetCommonLinkProperties(struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
        ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
                "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
                "<NewWANAccessType>Ethernet</NewWANAccessType><NewLayer1UpstreamMaxBitRate>100000000</NewLayer1UpstreamMaxBitRate><NewLayer1DownstreamMaxBitRate>100000000</NewLayer1DownstreamMaxBitRate><NewPhysicalLinkStatus>Up</NewPhysicalLinkStatus>",
                ca_event->ActionName);
        ca_event->ActionResult = UpnpParse_Buffer(result_str);

        return(ca_event->ErrCode);


}

int _GateDeviceGetTotalBytesSent(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	char dev[15];
	char *iface;
	FILE *stream;
	unsigned long bytes=0, total=0;

	/* Read sent from /proc */
	stream = fopen ( "/proc/net/dev", "r" );
	if ( stream != NULL )
	{
		iface=pGate->m_ipcon->IPCon_GetIfName(pGate->m_ipcon);
		while ( getc ( stream ) != '\n' );
		while ( getc ( stream ) != '\n' );

		while ( !feof( stream ) )
		{
			fscanf ( stream, "%[^:]:%*u %*u %*u %*u %*u %*u %*u %*u %lu %*u %*u %*u %*u %*u %*u %*u\n", dev, &bytes );
			if ( strcmp ( dev, iface )==0 )
				total += bytes;
		}
		fclose ( stream );
	}
	else
	{
		total=1;
	}

        ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n<NewTotalBytesSent>%lu</NewTotalBytesSent>\n</u:%sResponse>",
		ca_event->ActionName,
                "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
                total, ca_event->ActionName);
        ca_event->ActionResult = UpnpParse_Buffer(result_str);

        return(ca_event->ErrCode);

}

int _GateDeviceGetTotalBytesReceived(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	char dev[15];
        char *iface;
        FILE *stream;
        unsigned long bytes=0,total=0;

	/* Read received from /proc */
        stream = fopen ( "/proc/net/dev", "r" );
        if ( stream != NULL )
        {
                iface=pGate->m_ipcon->IPCon_GetIfName(pGate->m_ipcon);

                while ( getc ( stream ) != '\n' );
                while ( getc ( stream ) != '\n' );

                while ( !feof( stream ) )
                {
                        fscanf ( stream, "%[^:]:%lu %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u\n", dev, &bytes );
                        if ( strcmp ( dev, iface )==0 )
                                total += bytes;
                }
		fclose ( stream );
        }
	else
	{
                total=1;
        }

        ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n<NewTotalBytesReceived>%lu</NewTotalBytesReceived>\n</u:%sResponse>",
		ca_event->ActionName,
                "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
                total, ca_event->ActionName );
        ca_event->ActionResult = UpnpParse_Buffer(result_str);

        return(ca_event->ErrCode);

}

int _GateDeviceGetTotalPacketsSent(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	char dev[15];
        char *iface;
        FILE *stream;
        unsigned long pkt=0, total=0;

        /* Read sent from /proc */
        stream = fopen ( "/proc/net/dev", "r" );
        if ( stream != NULL )
        {
                iface=pGate->m_ipcon->IPCon_GetIfName(pGate->m_ipcon);
                while ( getc ( stream ) != '\n' );
                while ( getc ( stream ) != '\n' );

                while ( !feof( stream ) )
                {
                        fscanf ( stream, "%[^:]:%*u %*u %*u %*u %*u %*u %*u %*u %*u %lu %*u %*u %*u %*u %*u %*u\n", dev, &pkt );
                        if ( strcmp ( dev, iface )==0 )
                                total += pkt;
		}
                fclose ( stream );
        }
        else
        {
                total=1;
        }

        ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n<NewTotalPacketsSent>%lu</NewTotalPacketsSent>\n</u:%sResponse>", ca_event->ActionName,
                "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
                total, ca_event->ActionName);
        ca_event->ActionResult = UpnpParse_Buffer(result_str);

        return(ca_event->ErrCode);

}

 int _GateDeviceGetTotalPacketsReceived(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	char dev[15];
        char *iface;
        FILE *stream;
        unsigned long pkt=0, total=0;

        /* Read sent from /proc */
        stream = fopen ( "/proc/net/dev", "r" );
        if ( stream != NULL )
        {
                iface=pGate->m_ipcon->IPCon_GetIfName(pGate->m_ipcon);
                while ( getc ( stream ) != '\n' );
                while ( getc ( stream ) != '\n' );

                while ( !feof( stream ) )
                {
                        fscanf ( stream, "%[^:]:%*u %lu %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u\n", dev, &pkt );
                        if ( strcmp ( dev, iface )==0 )
                                total += pkt;
                }
                fclose ( stream );
        }
        else
	{
                total=1;
        }

        ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n<NewTotalPacketsReceived>%lu</NewTotalPacketsReceived>\n</u:%sResponse>", ca_event->ActionName,
                "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
                total, ca_event->ActionName);
        ca_event->ActionResult = UpnpParse_Buffer(result_str);

        return(ca_event->ErrCode);

}

int _GateDeviceSetConnectionType(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceGetConnectionTypeInfo(struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));

	ca_event->ErrCode = UPNP_E_SUCCESS;
	sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
		"urn:schemas-upnp-org:service:WANIPConnection:1",
		"<NewConnectionType>IP_Routed</NewConnectionType>\n<NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>",
		ca_event->ActionName);
	ca_event->ActionResult = UpnpParse_Buffer(result_str);

	return(ca_event->ErrCode);
}

int _GateDeviceRequestConnection(struct Upnp_Action_Request *ca_event)
{
#if 1 //lsshi 2003-8-13 11:54

	char result_str[500];
	
	
	ca_event->ErrCode = UPNP_E_SUCCESS;
	sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", ca_event->ActionName,
		"urn:schemas-upnp-org:service:WANIPConnection:1",
		ca_event->ActionName);
		
	ca_event->ActionResult = UpnpParse_Buffer(result_str);
	
#endif
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	return(UPNP_E_SUCCESS);
}

int _GateDeviceRequestTermination(struct Upnp_Action_Request *ca_event)
{
//	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceForceTermination(struct Upnp_Action_Request *ca_event)
{
	
#if 1 //lsshi 2003-8-15 14:12

	char result_str[500];
	
	
	ca_event->ErrCode = UPNP_E_SUCCESS;
	sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\"></u:%sResponse>", ca_event->ActionName,
		"urn:schemas-upnp-org:service:WANIPConnection:1",
		ca_event->ActionName);
		
	ca_event->ActionResult = UpnpParse_Buffer(result_str);
	
#endif	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceSetAutoDisconnectTime(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceSetIdleDisconnectTime(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceSetWarnDisconnectDelay(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}



int _GateDeviceGetStatusInfo(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{

	long int uptime;
	char result_str[500];
	
	uptime = (time(NULL) - pGate->startup_time);
	ca_event->ErrCode = UPNP_E_SUCCESS;
        sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n<NewConnectionStatus>%s</NewConnectionStatus><NewLastConnectionError>ERROR_NONE</NewLastConnectionError><NewUptime>%li</NewUptime>\n</u:%sResponse>", ca_event->ActionName,
	                "urn:schemas-upnp-org:service:WANIPConnection:1",
	                "Connected",uptime,
	                ca_event->ActionName);
        ca_event->ActionResult = UpnpParse_Buffer(result_str);
        return(ca_event->ErrCode);
}

int _GateDeviceGetAutoDisconnectTime(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceGetIdleDisconnectTime(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceGetWarnDisconnectDelay(struct Upnp_Action_Request *ca_event)
{
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	return(UPNP_E_SUCCESS);
}

int _GateDeviceGetNATRSIPStatus(struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));

	ca_event->ErrCode = UPNP_E_SUCCESS;
	sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
		"urn:schemas-upnp-org:service:WANIPConnection:1",
		"<NewRSIPAvailable>0</NewRSIPAvailable>\n<NewNATEnabled>1</NewNATEnabled>",
		ca_event->ActionName);
	ca_event->ActionResult = UpnpParse_Buffer(result_str);

	return(ca_event->ErrCode);
}

int _GateDeviceGetExternalIPAddress(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char result_str[500];
	char result_parm[500];
	char *ip_address = NULL;
	
	_UPNP_DEBUG(printf("%s %s %d\n",__FILE__,__FUNCTION__,__LINE__));
	
	ip_address =pGate->m_ipcon->IPCon_GetIpAddrStr(pGate->m_ipcon);	

	//printf("\nUpnp: GetExternalIPAddress=%s ", ip_address ); fflush(stdout);

	ca_event->ErrCode = UPNP_E_SUCCESS;
	sprintf(result_parm,"<NewExternalIPAddress>%s</NewExternalIPAddress>", ip_address);
	sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
		"urn:schemas-upnp-org:service:WANIPConnection:1",result_parm, ca_event->ActionName);
	ca_event->ActionResult = UpnpParse_Buffer(result_str);
	
	
	if (ip_address) free(ip_address);

	return(ca_event->ErrCode);
}

int _GateDeviceGetGenericPortMappingEntry(struct Upnp_Action_Request *ca_event)
{
	int i = 0;
	int index = 0;
	char *mapindex = NULL;
	int action_succeeded = 0;
	char result_parm[500];
	char result_str[500];
	char* address;
	
//lsshi add 2003-8-21 9:15
	address = gate.m_ipcon->IPCon_GetIpAddrStr(gate.m_ipcon);
	
	if(address)
	{
		printf("Now OUT UPNP IP is %s\n",address);
		
		free(address);
	}
	
	if ((mapindex = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewPortMappingIndex")))
	{
		
		Port_ListNode* node;
	
		index = atoi(mapindex);
		
#if 0		
		for (list<PortMap *>::iterator itr = m_list.m_pmap.begin(); itr != m_list.m_pmap.end(); itr++,i++)
		{
			if (i == index)
			{				
				sprintf(result_parm,"<NewRemoteHost>%s</NewRemoteHost><NewExternalPort>%d</NewExternalPort><NewProtocol>%s</NewProtocol><NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>%d</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%li</NewLeaseDuration>",
					"",(*itr)->m_ExternalPort, (*itr)->m_PortMappingProtocol, (*itr)->m_InternalPort,
				       	(*itr)->m_InternalClient, 1, (*itr)->m_PortMappingDescription, (*itr)->m_PortMappingLeaseDuration);		
				action_succeeded = 1;
			}
	
		}
#else
		
	
	
	node=PortListHead(&PORT_LIST);

	
	while(node)
	{
		if (i == index)
		{				
			sprintf(result_parm,"<NewRemoteHost>%s</NewRemoteHost><NewExternalPort>%d</NewExternalPort><NewProtocol>%s</NewProtocol><NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>%d</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%li</NewLeaseDuration>",
				"",PORT_ITEM->m_ExternalPort, PORT_ITEM->m_PortMappingProtocol, PORT_ITEM->m_InternalPort,
				 PORT_ITEM->m_InternalClient, 1, PORT_ITEM->m_PortMappingDescription, PORT_ITEM->m_PortMappingLeaseDuration);		
			action_succeeded = 1;
		}
		
		node=node->next;
		i++;
		
	
	}


#endif
		
		
		if (action_succeeded)	//lsshi 2003-8-13 13:46 if (action_succeeded)
		{
			
			ca_event->ErrCode = UPNP_E_SUCCESS;
			sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
	                        "urn:schemas-upnp-org:service:WANIPConnection:1",result_parm, ca_event->ActionName);
			ca_event->ActionResult = UpnpParse_Buffer(result_str);
		}
		else
		{
			
			ca_event->ErrCode = 713;
			strcpy(ca_event->ErrStr, "SpecifiedArrayIndexInvalid");
			ca_event->ActionResult = NULL;
		}

	}
	else
	{
			SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceGetGenericortMappingEntry: Invalid Args");
			ca_event->ErrCode = 402;
			strcpy(ca_event->ErrStr, "Invalid Args");
			ca_event->ActionResult = NULL;
	}

	if (mapindex) free(mapindex);
	
	
	return(ca_event->ErrCode);
}

void PrintList(PortLinkedList * list);

int _GateDeviceGetSpecificPortMappingEntry(struct Upnp_Action_Request *ca_event)
{
	char *ext_port=NULL;
	char *proto=NULL;
	char result_parm[500];
	char result_str[500];
	int prt;
	int action_succeeded = 0;
	
	if ((ext_port = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort")) 
		&& (proto = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest,"NewProtocol")))
	{
		prt = getProtoNum(proto);
		if ((prt == 17) || (prt == 6))
		{
			
			PrintList(&PORT_LIST);
			
			printf("List size is %d\n",PortListSize(&PORT_LIST));
			
#if 0
			for (list<PortMap *>::iterator itr = m_list.m_pmap.begin(); itr != m_list.m_pmap.end(); itr++)
			{
				if ((((*itr)->m_ExternalPort == atoi(ext_port)) && (strcmp((*itr)->m_PortMappingProtocol,proto) == 0)))
				{
					sprintf(result_parm,"<NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>1</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%li</NewLeaseDuration>",
							(*itr)->m_InternalPort, 
							(*itr)->m_InternalClient,
							(*itr)->m_PortMappingDescription,
							(*itr)->m_PortMappingLeaseDuration);
					action_succeeded = 1;
				}
			}
			
#else

			Port_ListNode* node;
	
			node=PortListHead(&PORT_LIST);
	
		
			while(node)
			{
				if (((PORT_ITEM->m_ExternalPort == atoi(ext_port)) && (strcmp(PORT_ITEM->m_PortMappingProtocol,proto) == 0)))
				{
					sprintf(result_parm,"<NewInternalPort>%d</NewInternalPort><NewInternalClient>%s</NewInternalClient><NewEnabled>1</NewEnabled><NewPortMappingDescription>%s</NewPortMappingDescription><NewLeaseDuration>%li</NewLeaseDuration>",
							PORT_ITEM->m_InternalPort, 
							PORT_ITEM->m_InternalClient,
							PORT_ITEM->m_PortMappingDescription,
							PORT_ITEM->m_PortMappingLeaseDuration);
							
					action_succeeded = 1;
				}
				
				node=node->next;
				
					
			}
			
#endif
		
			if (action_succeeded)
			{
				
				ca_event->ErrCode = UPNP_E_SUCCESS;
		                sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>", ca_event->ActionName,
		                        "urn:schemas-upnp-org:service:WANIPConnection:1",result_parm, ca_event->ActionName);
		                ca_event->ActionResult = UpnpParse_Buffer(result_str);
			}
			else
			{
				SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceGetSpecificPortMappingEntry: PortMapping Doesn't Exist...");
		                ca_event->ErrCode = 714;
		                strcpy(ca_event->ErrStr, "NoSuchEntryInArray");
		                ca_event->ActionResult = NULL;
			}
		}
		else
		{
		        SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceGetSpecificPortMappingEntry: Invalid NewProtocol=%s\n",proto);
                        ca_event->ErrCode = 402;
                        strcpy(ca_event->ErrStr, "Invalid Args");
                        ca_event->ActionResult = NULL;
		}
	}
	else
	{
		SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceGetSpecificPortMappingEntry: Invalid Args");
		ca_event->ErrCode = 402;
		strcpy(ca_event->ErrStr, "Invalid Args");
		ca_event->ActionResult = NULL;
	}
		
	return (ca_event->ErrCode);
}

int _GateDeviceAddPortMapping(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char *remote_host=NULL;
	char *ext_port=NULL;
	char *proto=NULL; 
	char *int_port=NULL;
	char *int_ip=NULL;
	char *desc=NULL;
	int prt,result=0;
	char num[5];
	char result_str[500];
	char *address = NULL;
	Upnp_Document  PropSet= NULL;	
	int action_succeeded = 0;
		
	address = pGate->m_ipcon->IPCon_GetIpAddrStr(pGate->m_ipcon);

	remote_host = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewRemoteHost");

	if (
		(ext_port = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort"))
		&& (proto    = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest,"NewProtocol"))
		&& (int_port = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewInternalPort"))
		&& (int_ip   = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewInternalClient"))
		&& (desc     = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewPortMappingDescription")))
	{		
		prt = getProtoNum(proto);
		if ((prt == 6) || (prt ==17) )
		{
			if (chkIPADDRstring(int_ip)!=0)
			{
#if 0
				for (list<PortMap *>::iterator itr=m_list.m_pmap.begin(); itr != m_list.m_pmap.end(); itr++)
				{
					if (((*itr)->m_ExternalPort == atoi(ext_port)) 
						&& (strcmp((*itr)->m_PortMappingProtocol,proto) == 0) 
						&& (strcmp((*itr)->m_InternalClient,int_ip) == 0))
					{
						m_list.delPortForward((*itr)->m_PortMappingProtocol, 
								(*itr)->m_ExternalIP,
								(*itr)->m_ExternalPort, (*itr)->m_InternalClient,
								(*itr)->m_InternalPort);
						delete *itr;
						m_list.m_pmap.erase(itr);
						itr--;
					}
				}
				
#else

				Port_ListNode* node;
				Port_ListNode* del;
		
				node=PortListHead(&PORT_LIST);
					
				while(node)
				{
					if ((PORT_ITEM->m_ExternalPort == atoi(ext_port)) 
						&& (strcmp(PORT_ITEM->m_PortMappingProtocol,proto) == 0) 
						&& (strcmp(PORT_ITEM->m_InternalClient,int_ip) == 0) )
					{
						if( remote_host && strcmp(PORT_ITEM->m_RemoteHost,remote_host)!=0 )
						{
							node=node->next;
							continue;
						}
						
						pGate->m_list.delPortForward(PORT_ITEM->m_RemoteHost, PORT_ITEM->m_PortMappingProtocol, 
								PORT_ITEM->m_ExternalIP,
								PORT_ITEM->m_ExternalPort, PORT_ITEM->m_InternalClient,
								PORT_ITEM->m_InternalPort);
						
						del=node;
						node=node->next;
						
						PortListDelNode(&PORT_LIST,del,1);
						//free(PORT_ITEM);
						
					}
					else
						node=node->next;
					
				
				}					

#endif

				result=pGate->m_list.PortMapAdd(remote_host, proto, address, atoi(ext_port), int_ip, atoi(int_port), 1, desc, 0);
				if (result==1)
				{
					sprintf(num,"%d",PortListSize(&PORT_LIST));
					PropSet= UpnpCreatePropertySet(1,"PortMappingNumberOfEntries", num);
					UpnpNotifyExt(pGate->device_handle, ca_event->DevUDN,ca_event->ServiceID,PropSet);
					UpnpDocument_free(PropSet);
					//SYS_DEBUG(LOG_DEBUG, "AddPortMap: RemoteHost: %s Prot: %d ExtPort: %d Int: %s.%d\n-----------------------\n",
					//	remote_host, prt, atoi(ext_port), int_ip, atoi(int_port));
					action_succeeded = 1;
				}
				else
				{
					if (result==718)
					{
						SYS_DEBUG(LOG_DEBUG,"Failure in GateDeviceAddPortMapping: RemoteHost: %s Prot:%d ExtPort: %d Int: %s.%d\n",
							remote_host,prt, atoi(ext_port),int_ip, atoi(int_port));
						ca_event->ErrCode = 718;
						strcpy(ca_event->ErrStr, "ConflictInMappingEntry");
						ca_event->ActionResult = NULL;
					}
				}
			}
			else
			{
		                 SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceAddPortMapping: Invalid NewInternalClient=%s\n",int_ip);
                                 ca_event->ErrCode = 402;
                                 strcpy(ca_event->ErrStr, "Invalid Args");
                                 ca_event->ActionResult = NULL;
			}
		}
		else
		{
		      SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceAddPortMapping: Invalid NewProtocol=%s\n",proto);
                      ca_event->ErrCode = 402;
                      strcpy(ca_event->ErrStr, "Invalid Args");
                      ca_event->ActionResult = NULL;
		}
	}
	else
	{
		SYS_DEBUG(LOG_DEBUG, "Failiure in GateDeviceAddPortMapping: Invalid Arguments!");
                ca_event->ErrCode = 402;
                strcpy(ca_event->ErrStr, "Invalid Args");
                ca_event->ActionResult = NULL;
	}
	
	if (action_succeeded)
	{
		ca_event->ErrCode = UPNP_E_SUCCESS;
		sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>",
			ca_event->ActionName, "urn:schemas-upnp-org:service:WANIPConnection:1", "", ca_event->ActionName);
		ca_event->ActionResult = UpnpParse_Buffer(result_str);
	}

	if (ext_port) free(ext_port);
	if (int_port) free(int_port);
	if (proto) free(proto);
	if (int_ip) free(int_ip);
	if (desc) free(desc);
	if (remote_host) free(remote_host);	
	if (address) free(address);
	
	return(ca_event->ErrCode);
}

int _GateDeviceDeletePortMapping(struct _Gate* pGate,struct Upnp_Action_Request *ca_event)
{
	char *remote_host=NULL;
	char *ext_port=NULL;
	char *proto=NULL;
	int prt,result=0;
	char num[5];
	char result_str[500];
	Upnp_Document  PropSet= NULL;
	int action_succeeded = 0;

	remote_host= SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewRemoteHost") ;
	
	if ((
		(ext_port = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewExternalPort")) &&
		(proto = SampleUtil_GetFirstDocumentItem(ca_event->ActionRequest, "NewProtocol"))))
	{

		prt = getProtoNum(proto);
		if ((prt == 6) || (prt == 17))
		{
			result=pGate->m_list.PortMapDelete(remote_host, proto, atoi(ext_port));
			if (result==1)
			{
				sprintf(num,"%d",PortListSize(&PORT_LIST));
				
				PropSet= UpnpCreatePropertySet(1,"PortMappingNumberOfEntries", num);
				UpnpNotifyExt(pGate->device_handle, ca_event->DevUDN,ca_event->ServiceID,PropSet);
				UpnpDocument_free(PropSet);
				action_succeeded = 1;
			}
			else
			{
				SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceDeletePortMapping: DeletePortMap: RemoteHost: %s Proto:%s Port:%s\n", remote_host, proto, ext_port);
				ca_event->ErrCode = 714;
				strcpy(ca_event->ErrStr, "NoSuchEntryInArray");
				ca_event->ActionResult = NULL;
			}
		}
		else
		{
			SYS_DEBUG(LOG_DEBUG, "Failure in GateDeviceDeletePortMapping: Invalid NewProtocol=%s\n",proto);
			ca_event->ErrCode = 402;
                        strcpy(ca_event->ErrStr, "Invalid Args");
                        ca_event->ActionResult = NULL;
		}
	}
	else
	{
	        SYS_DEBUG(LOG_DEBUG, "Failiure in GateDeviceDeletePortMapping: Invalid Arguments!");
       	        ca_event->ErrCode = 402;
       	        strcpy(ca_event->ErrStr, "Invalid Args");
	        ca_event->ActionResult = NULL;
	}

	if (action_succeeded)
	{
		ca_event->ErrCode = UPNP_E_SUCCESS;
		sprintf(result_str, "<u:%sResponse xmlns:u=\"%s\">\n%s\n</u:%sResponse>",
			ca_event->ActionName, "urn:schemas-upnp-org:service:WANIPConnection:1", "", ca_event->ActionName);
		ca_event->ActionResult = UpnpParse_Buffer(result_str);
	}

	if (ext_port) free(ext_port);
	if (proto) free(proto);
	
	return(ca_event->ErrCode);
}

