///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2003 C++ to C Made by PC32 Lsshi
// All rights reserved. 

///////////////////////////////////////////////////////////////////////////
#include "gate.h"
#include "main.h"
#include "sample_util.h"
#include "portmap.h"
#include "ipcon.h"
#include <stdio.h>
#include <asm/unistd.h>
#include <arpa/inet.h>
#include <linux/sockios.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>

/* return network addr in network order */
#define MAX_IFS	100
#define SIN_ADDR(x)     (((struct sockaddr_in *) (&(x)))->sin_addr.s_addr)

unsigned long getIfAddr (char *ifname )
{
    struct ifreq *ifr, *ifend;
    u_int32_t ina;
    struct ifconf ifc;
    struct ifreq ifs[MAX_IFS];
    int sockfd = -1;
    unsigned long ipaddr = 0;
	
    sockfd = socket(AF_INET, SOCK_DGRAM, 0);
    if (sockfd < 0)
	return 0;

    ifc.ifc_len = sizeof(ifs);
    ifc.ifc_req = ifs;
    if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0) {
	perror("ioctl(SIOCGIFCONF)" );
	return 0;
    }
/*
 * Scan through looking for an interface with an Internet
 * address on the same subnet as `ipaddr'.
 */
    ifend = ifs + (ifc.ifc_len / sizeof(struct ifreq));
    for (ifr = ifc.ifc_req; ifr < ifend; ifr++) {
	if (ifr->ifr_addr.sa_family == AF_INET) {
	    ina = SIN_ADDR(ifr->ifr_addr);
	    if( strcmp(ifr->ifr_name,ifname)!=0 )
	    	continue;
	    ipaddr = ina;
	    break;
	}
    }
    close(sockfd);
    return ipaddr;
}



void Destroy_Rules(void);

void C_IPCon_Init(struct _IPCon* pIPCon)
{
	
	pIPCon->IPCon_Void_Init=_IPCon_Void_Init;
	pIPCon->IPCon_Init=_IPCon_Init;
	pIPCon->IPCon_Void_Destroy=_IPCon_Void_Destroy;
	pIPCon->IPCon_SetIfName=_IPCon_SetIfName;
	pIPCon->IPCon_GetIfName=_IPCon_GetIfName;
	pIPCon->IPCon_GetIpAddrStr=_IPCon_GetIpAddrStr;
	pIPCon->IPCon_IsIfUp=_IPCon_IsIfUp;
	pIPCon->IPCon_GetIfStatStr=_IPCon_GetIfStatStr;
	//memset(pIPCon->Ext_IP,0,16);
	
	
}

int get_sockfd()
{
	static int sockfd = -1;
	
	if (sockfd == -1)
	{
		if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		{
			perror("user: socket creating failed");
			return (-1);
		}
	}
	return sockfd;
}

void _IPCon_Void_Init(struct _IPCon* pIPCon)
{
	pIPCon->m_ifname = (char*)malloc(18);
	//memset(pIPCon->Ext_IP,0,16);
}

void _IPCon_Init(struct _IPCon* pIPCon,char *ifname)
{
	pIPCon->m_ifname = (char*)malloc(strlen(ifname)+1);
	strcpy (pIPCon->m_ifname,ifname);
	//memset(pIPCon->Ext_IP,0,16);
	pIPCon->old_out_ip=0;
	pIPCon->now_out_ip=0;
}

void _IPCon_Void_Destroy(struct _IPCon* pIPCon)
{
	if (pIPCon->m_ifname)
	{
		free(pIPCon->m_ifname);
		pIPCon->m_ifname = NULL;
	}
	
	//memset(pIPCon->Ext_IP,0,16);
}

int _IPCon_SetIfName(struct _IPCon* pIPCon,char * ifname)
{
	if (strlen(ifname) < sizeof(pIPCon->m_ifname)+1)
	{
		strcpy(pIPCon->m_ifname,ifname);
		return 1;
	}
	else
	{
		printf("Something wrong with interface name !\n");
		pIPCon->m_ifname=NULL;
		return 0;
	}
}

char * _IPCon_GetIfName(struct _IPCon* pIPCon)
{
	return pIPCon->m_ifname;
}

char * _IPCon_GetIpAddrStr(struct _IPCon* pIPCon)
{
	struct ifreq ifr;
	struct sockaddr_in *saddr;
	int fd;
	char *address=NULL;
	int ret;
	unsigned long ip=0;

	address = (char*)malloc(20);
	strcpy(address,"0.0.0.0");
	
	if( pIPCon->m_ifname && pIPCon->m_ifname[0] )
	{
		if( (ip=getIfAddr(pIPCon->m_ifname) ) )
		{
			struct in_addr saddr;
			saddr.s_addr = ip;
			strcpy(address,inet_ntoa(saddr));
		}
	}

	return address;

}

int _IPCon_IsIfUp(struct _IPCon* pIPCon)
{
	struct ifreq ifr;
	int fd;

	fd = get_sockfd();
	if (fd >=0)
	{
		if (strlen(pIPCon->m_ifname) < sizeof(ifr.ifr_name))
			strcpy (ifr.ifr_name,pIPCon->m_ifname);
		ifr.ifr_addr.sa_family = AF_INET;
		if (ioctl(fd, SIOCGIFFLAGS, &ifr) < 0)
		{
			close(fd);
			return (-1);
		}
		if (ifr.ifr_flags & IFF_UP)
		{
			close(fd);
			return (1);
		}
		else
		{
			close(fd);
			return(0);
		}
	}
	return (-1);
}

char * _IPCon_GetIfStatStr(struct _IPCon* pIPCon)
{
	if (pIPCon->IPCon_IsIfUp(pIPCon))
		return ("UP");
	else
		return NULL;
}

