#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/if_ether.h>
#include <netinet/tcp.h>

#include "notify.h"

#define  SOCKET int
#define  SOCKADDR   sockaddr
#define  TRUE	1
#define  FALSE	0

#define  SOCKET_ERROR   -1
#define  INVALID_SOCKET -1

char *event_str[5]={"DHCP_UP","PPPOE_UP","DHCP_DOWN","PPPOE_DOWN", "UPNP_RESTART"};

int createClient( const char *hostname, int port )
{ 
    struct	sockaddr_in sa ; 
    struct	hostent *hp ; 
    SOCKET	s ; 
    long	addr ; 
	
    memset(&sa, 0, sizeof(sa)) ; 
    if ((addr = inet_addr(hostname)) != INADDR_NONE) { 
        memcpy( (char *) &sa.sin_addr,&addr, sizeof(addr)) ; 
        sa.sin_family = AF_INET ; 
    } 
	else { 
        if ((hp = gethostbyname(hostname)) == NULL) { 
            return INVALID_SOCKET ; 
       } 
        hostname = hp->h_name ; 
        memcpy((char *) &sa.sin_addr,hp->h_addr,  hp->h_length) ; 
        sa.sin_family = hp->h_addrtype ; 
    } 

    sa.sin_port = htons((u_short) port) ;    
    if ( (s = socket(AF_INET, SOCK_STREAM, 0)) == INVALID_SOCKET ) { 
		return INVALID_SOCKET ; 
    } 
 
    if ( connect(s, (struct sockaddr *)&sa, sizeof(sa)) == SOCKET_ERROR ) 
	{ 
        close(s) ;
        return SOCKET_ERROR ; 
    } 

    return s ;
} 

int writen( SOCKET fd, const char *ptr, int nbytes )
{
    int         nleft, nwritten; 
 
    nleft = nbytes; 
    while( nleft > 0 )
    {
        nwritten = send( fd, ptr, nleft, 0 );
        if( nwritten == 0 ) return 0;
        if( nwritten > 0 ) {
            nleft -= nwritten;
            ptr += nwritten;
        }
        else if( errno != EWOULDBLOCK && errno != ENOBUFS )
                return nwritten;
    } 
    return (nbytes - nleft);   

}


int notifyHttp(int event, char *ifname, int sessid )
{
	int sock =0, ret= 0;
	char buf[255];
	
	if( event >= EVENT_UNKNOWN )
		return -1;
	
	sock = createClient("127.0.0.1", 80 );
	if( sock != INVALID_SOCKET )
	{
		snprintf(buf,sizeof(buf)-1, "%s %s %d\r\n",  event_str[ event ], ifname ? ifname : "", sessid );
		ret = writen( sock, buf, strlen(buf) );
		close(sock);
	}
	
	return ret;
}

/*

int main(void)
{
	notifyHttp( EVENT_DHCP_UP );
	notifyHttp( EVENT_PPPOE_DOWN );
}
*/
