///////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2003 C++ to C Made by PC32 Lsshi
// All rights reserved. 

///////////////////////////////////////////////////////////////////////////
#include "portmap.h"
#include "portlist.h"
#include <malloc.h>

void C_PortMap_Init(struct _PortMap* pMap);
void _PortMap_Void_Init(struct _PortMap* pMap);
/****************************************************************************
 * Function: PortCreateListNode
 *
 *  Description:
 *      Creates a list node. Dynamically.
 *      
 *  Parameters:
 *      void * item - the item to store
 *  Returns:
 *      The new node, NULL on failure.
 *****************************************************************************/
static Port_ListNode *
PortCreateListNode( void *item )
{

    Port_ListNode *temp = NULL;


    temp = ( Port_ListNode * ) malloc(sizeof(Port_ListNode));
    
    C_PortMap_Init((struct _PortMap*)item);
    ((struct _PortMap*)item)->PortMap_Void_Init((struct _PortMap*)item);
    
    if( temp ) {
        temp->next = NULL;
        temp->item = item;
    }
    return temp;
}

/****************************************************************************
 * Function: PortListInit
 *
 *  Description:
 *      Initializes PortLinkedList. Must be called first.
 *      And only once for List.
 *  Parameters:
 *      list  - must be valid, non null, pointer to a linked list.
  *      free_func - function used to free items. (May be NULL)
 *  Returns:
 *      0 on success, EOUTOFMEM on failure.
 *****************************************************************************/
int
PortListInit( PortLinkedList * list,
                port_free_function free_func )
{
  //  int retCode = 0;
    if( list == NULL )
        return LIST_FAIL;

    list->size = 0;
    list->free_func = free_func;

  
    list->head= NULL;
    list->tail=NULL;

    return LIST_SUCCESS;
}


/****************************************************************************
 * Function: PortListAddTail
 *
 *  Description:
 *      Adds a node to the tail of the list.
 *      Node gets added immediately before list.tail.
 *  Parameters:
 *      PortLinkedList *list  - must be valid, non null, pointer to a linked list.
 *      void * item - item to be added
 *  Returns:
 *      The pointer to the Port_ListNode on success, NULL on failure.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
Port_ListNode *
PortListAddTail( PortLinkedList * list,
             void *item )
{
 
	Port_ListNode *temp;

    if( list == NULL )
        return NULL;
  
    	
    temp=PortCreateListNode(item);

    if(list->head==NULL)
    {
    	list->head=temp;
    	temp->next=NULL;
    	list->tail=temp;
		list->size++;
    }
    else
    {
    	
    	list->tail->next=temp;
    	list->tail=temp;
	temp->next=NULL;
	list->size++;
    	
    	
    }

	return list->tail;
}



/****************************************************************************
 * Function: PortListDelNode
 *
 *  Description:
 *      Removes a node from the list
 *      The memory for the node is freed but the
 *      the memory for the items are not.
 *  Parameters:
 *      PortLinkedList *list  - must be valid, non null, pointer to a linked list.
 *      Port_ListNode *dnode - done to delete.
 *  Returns:
 *      The pointer to the item stored in node on success, NULL on failure.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int
PortListDelNode( PortLinkedList * list,
             Port_ListNode * dnode,
             int freeItem )
{
    Port_ListNode *cur=NULL;
    Port_ListNode *prev=NULL;
    
    if((list == NULL )||(dnode==NULL))
    {
    	return LIST_FAIL;
    }
  
   for(cur=list->head;cur;cur=cur->next)
   {
   	   	
   	if(cur==dnode)
   	{   		
   	    if(cur==list->head)
	    {
	    	list->head=cur->next;

	    	if(freeItem)
	    	{
	    		if(cur->item)
	    		{
	    			list->free_func(cur->item);
	    			free(cur->item);
	    		}
	    	}
	    	
		free(cur);
	    	list->size--;
	    	
	    	if(list->head && list->head->next==NULL)
	    		list->tail=list->head;

	    	return LIST_SUCCESS;
	    	
	    }//head
	    else
	    {
	    	
	    	prev->next=cur->next;

		printf("PortListDelNode 2\n");fflush(stdout);

	    	if(freeItem)
	    	{
	    		list->free_func(cur->item);	    		
	    		free(cur->item);	    	
	    	}
	    	
	    	free(cur);
	    	list->size--;
	    	
	    	if(prev->next==NULL)
	    		list->tail=prev;
	    		
	    	return LIST_SUCCESS;
	    	
	    }
	
   	}
   	
   	prev=cur;
   	
   }
   
   return LIST_FAIL;

}

/****************************************************************************
 * Function: PortListDestroy
 *
 *  Description:
 *      Removes all memory associated with list nodes. 
 *      Does not free PortLinkedList *list. 
 *      Items stored in the list are not freed, only nodes are.
 *  Parameters:
 *      PortLinkedList *list  - must be valid, non null, pointer to a linked list.
 *  Returns:
 *      0 on success. Nonzero on failure.
 *      Always returns 0.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int
PortListDestroy( PortLinkedList * list,
             int freeItem )
{
    Port_ListNode *cur=NULL;
	Port_ListNode *now=NULL;
  
    if(list == NULL )
    {
    	return LIST_FAIL;
    }
  
   for(cur=list->head;cur;)
   {
		now=cur;

	    	if(freeItem)
	    	{
	    		list->free_func(now->item);
	    		free(now->item);
	    	
	    	}
	    
		cur=cur->next;
	    	free(now);
	
   }
   
   list->head=NULL;
   list->tail=NULL;
   list->free_func=NULL;
   list->size=0;

   return LIST_SUCCESS;
   
}

/****************************************************************************
 * Function: ListHead
 *
 *  Description:
 *      Returns the head of the list.
 *    
 *  Parameters:
 *      PortLinkedList *list  - must be valid, non null, pointer to a linked list.
 *  
 *  Returns:
 *      The head of the list. NULL if list is empty.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
Port_ListNode *
PortListHead( PortLinkedList * list )
{
   
    if( list == NULL )
        return NULL;

    if( list->size == 0 )
        return NULL;
    else
        return list->head;
}




/****************************************************************************
 * Function: ListSize
 *
 *  Description:
 *     Returns the size of the list.
 *  Parameters:
 *      PortLinkedList *list  - must be valid, non null, pointer to a linked list.
 
 *  Returns:
 *      The number of items in the list.
 *  Precondition:
 *      The list has been initialized.
 *****************************************************************************/
int
PortListSize( PortLinkedList * list )
{
   
    if( list == NULL )
        return LIST_FAIL;

    return list->size;
}


#include "portmap.h"

void PrintList(PortLinkedList * list)
{
	Port_ListNode* node;
	
	struct _PortMap* pMap;
		
	node=PortListHead(list);
	
	while(node)
	{
		pMap=(struct _PortMap*)node->item;
		
	        //printf("%s \n",pMap->m_RemoteHost);
	        printf("pMap->m_ExternalIP %s \n",pMap->m_ExternalIP);
	        printf("pMap->m_InternalClient %s \n",pMap->m_InternalClient);
	        printf("pMap->m_PortMappingDescription %s \n",pMap->m_PortMappingDescription);
	        
	        printf("pMap->m_PortMappingProtocol %s \n",pMap->m_PortMappingProtocol);
	        printf("pMap->m_ExternalPort %d \n",pMap->m_ExternalPort);
      
		node=node->next;
		
	}
	
}
