/* $Id: netif_mgmt.h,v 1.1 2008/09/22 09:13:46 bafchang Exp $ */

/**
 * @file netif_mgmt.h
 *
 *
 */

/* Copyright (c) 2000-2007 ASUSTek Computer, Inc. All rights reserved.
 *
 * ASUSTek Computer, Inc. Confidential and Proprietary
 *
 * The following software source code ("Software") is strictly confidential and
 * is proprietary to ASUSTek Computer, Inc. ("ASUSTek").  It may only be read,
 * used, copied, adapted, modified or otherwise dealt with by you if you have
 * entered into a confidentiality agreement with ASUSTek and then subject to the
 * terms of that confidentiality agreement and any other applicable agreement
 * between you and ASUSTek.  If you are in any doubt as to whether you are
 * entitled to access, read, use, copy, adapt, modify or otherwise deal with
 * the Software or whether you are entitled to disclose the Software to any
 * other person you should contact ASUSTek.  If you have not entered into a
 * confidentiality agreement with ASUSTek granting access to this Software you
 * should forthwith return all media, copies and printed listings containing
 * the Software to ASUSTek.
 *
 * ASUSTek reserves the right to take legal action against you should you breach
 * the above provisions.
 */

/*-------------------------------------------------------------------------
Modification History 
-------------------- 
$Log: netif_mgmt.h,v $
Revision 1.1  2008/09/22 09:13:46  bafchang
Added pptp-1.7.0 and xl2tpd-1.1.12 into the CVS repository.

Revision 1.1.2.1  2008/04/18 07:59:37  travis
bind socket with wan ipaddress

Revision 1.1.2.2  2008/03/14 11:03:23  travis
features modification

*/

#ifndef _NETIF_H_
#define _NETIF_H_

#include <sys/types.h>

/* netif_mgmt kernel module required data are specified here  */

/* NETIF socket options */
#define NETIF_BASE_CTL		(64+1024+64)		/* base */

/* SET */
#define NETIF_SO_SET_NONE	NETIF_BASE_CTL	        /* just peek */
#define NETIF_SO_SET_LINK	(NETIF_BASE_CTL+1)

#define NETIF_SO_SET_DEL	(NETIF_BASE_CTL+2)
#define NETIF_SO_SET_ATTACH	(NETIF_BASE_CTL+3)
#define NETIF_SO_SET_DETACH	(NETIF_BASE_CTL+4)
#define NETIF_SO_SET_SYNC	(NETIF_BASE_CTL+5)
#define NETIF_SO_SET_CLEAR	(NETIF_BASE_CTL+6)
#define NETIF_SO_SET_LB		(NETIF_BASE_CTL+7)
#define NETIF_SO_SET_MAX	(NETIF_BASE_CTL+8)

/* GET */
#define NETIF_SO_GET_LINK	(NETIF_BASE_CTL+14)
#define NETIF_SO_GET_LINKS	(NETIF_BASE_CTL+15)
#define NETIF_SO_GET_MAX	(NETIF_BASE_CTL+16)

#define NETIF_IFNAME_MAXLEN	16

#define MAX_INTERFACES	8

#define KBPS		(1024)					/* Kilo BITs per second */
#define MBPS		(1024 * 1024)			/* Mega BITs per second */
#define GBPS		(1024 * 1024 * 1024)	/* Giga BITs per second */
#define	MAX_BW		(1 * GBPS)

#define MAX_LB_LINK	16

#define LB_LINK_START	4	/* eth0.2 (WAN1) */
#define LB_LINK_END	7	/* eth0.5 (WAN4) */

#define LB_TYPE_LAN	0
#define LB_TYPE_WAN	1

enum lb_op { 
	LB_OP_DISABLE = 0, 
	LB_OP_LOADBALANCE, 
	LB_OP_BACKUP
};
	
enum lb_algo { 
	LB_WRR = 1, 
	LB_LTF 
};

enum netif_opcode { 
	NETIF_OP_NONE = 0x0,
	NETIF_OP_LINK = 0x1,
	NETIF_OP_STATUS = 0x2, 
	NETIF_OP_GW = 0x4, 
	NETIF_OP_PPPOE = 0x8, 
	NETIF_OP_BW = 0x10, 
	NETIF_OP_CHECK = 0x20,
	NETIF_OP_WEIGHT = 0x40
};

enum netif_status { 
	NETIF_STATUS_OFF = 0,
	NETIF_STATUS_ON = 1
};

struct netif_stats
{
	uint64_t	rx_bytes;		/* total bytes received 	*/
	uint64_t	tx_bytes;		/* total bytes transmitted	*/

	uint64_t	rx_bytes_last;		/* last total bytes received */
	uint64_t	tx_bytes_last;		

	uint64_t	rx_bytes_avg;		/* average 	*/
	uint64_t	tx_bytes_avg;	

	uint64_t	rx_load_avg;		/* load	*/
	uint64_t	tx_load_avg;

	uint64_t	rx_bytes_max;
	uint64_t	tx_bytes_max;

	uint64_t	cur_avail_bw_up;	/* cur avg bw up */
	uint64_t	cur_avail_bw_down;	/* cur avg bw down */
};

struct netif_link {

	char name[NETIF_IFNAME_MAXLEN];
	char ppp_name[NETIF_IFNAME_MAXLEN];	/* attached ppp name */

	uint16_t	type;		/* lan: 0 , wan: 1 */
	uint16_t	ifnum;		/* ifnum */

	uint16_t	phy_ifnum;	/* physical ifnum */
	uint16_t	ppp_ifnum;	/* attached ppp ifnum */
	uint16_t	pppoe_id;	/* pppoe id */
	unsigned char	loc_mac[6];	/* use for pppoe */
	unsigned char	peer_mac[6];	/* use for pppoe */
	
	uint16_t	cur_weight;	/* for weighted round robin */
	uint16_t	max_weight;	/* for weighted round robin */

	uint64_t	bw_up;		/* upload kbps */
	uint64_t	bw_down;	/* download kbps */
	uint32_t	ip;
	uint32_t	mask;
	uint32_t	network;
	uint32_t	gw[2];		/* two gateway */
	uint32_t	testip;		/* test ip */

	uint16_t	status[2];		/* up or down , two pppoe per interface */
	uint16_t	primary;	/* for backup mode only */
	
	uint16_t	ok;		/* use by netif_lb_link_sync() : to remove obsoleted interface */		

	uint64_t	uptime[2];		/* uptime */

	/* cache data from dev_base */
	struct netif_stats	stats;

};


struct netif_rule_user {

	char name[NETIF_IFNAME_MAXLEN];
	char ppp_name[NETIF_IFNAME_MAXLEN];	/* attached ppp name */

	uint16_t	type;		/* lan: 0 , wan: 1 */
	uint16_t	ifnum;		/* ifnum */

	/* pppoe fastpath */
	uint16_t	phy_ifnum;	/* physical ifnum */
	uint16_t	ppp_ifnum;	/* attached ppp ifnum */
	uint16_t	pppoe_id;	/* pppoe id */
	unsigned char	loc_mac[6];	/* use for pppoe */
	unsigned char	peer_mac[6];	/* use for pppoe */
	
	uint16_t	cur_weight;	/* for weighted round robin */
	uint16_t	max_weight;	/* for weighted round robin */

	uint64_t	bw_up;		/* upload kbps */
	uint64_t	bw_down;	/* download kbps */
	uint32_t	ip;
	uint32_t	mask;
	uint32_t	network;
	uint32_t	gw[2];
	uint32_t	testip;		/* test ip */

	uint16_t	status[2];		/* up or down */
	uint16_t	primary;	/* for backup mode only */

	uint16_t	ok;		/* use by netif_lb_link_sync() : to remove obsoleted interface */	

	uint64_t	uptime[2];		/* uptime */

	uint16_t  opcode;
	uint16_t  param;
	uint16_t  operation;	/* load_balance: 0, backup: 1, disable: 2 */
	char	primary_link[NETIF_IFNAME_MAXLEN];	/* primary link when operation is backup type */

};
#endif
