#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>

#include "netif_mgmt_util.h"

int netif_get_iflist(char *buf, int size)
{
	struct netif_rule_user urule;
	int cmd;
	int len = sizeof(struct netif_rule_user);
	int sockfd, result = 0;
		
	cmd = NETIF_SO_GET_LINKS;

	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
                return -1;
                
	result = getsockopt(sockfd, IPPROTO_IP,	cmd, &urule, (socklen_t *)&len);
	
	strncpy(buf, (char *)&urule, size > sizeof(struct netif_rule_user) ? sizeof(struct netif_rule_user) : size-1);

	close(sockfd);
	
	return 0;		
}

int netif_get_info(char *name, struct netif_rule_user *urule )
{
	int cmd;
	int len = sizeof(struct netif_rule_user);
	
	int sockfd;
	int result = 0;

	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
                return -1;
                		
	cmd = NETIF_SO_GET_LINK;
	strncpy(urule->name, name, sizeof(urule->name)-1);
	
	result = getsockopt(sockfd, IPPROTO_IP,	cmd, urule, (socklen_t *)&len);
	
	close(sockfd);

	return result;
}

int netif_get_whole_info(char *name, struct netif_rule_user *urule )
{
	struct netif_rule_user ppp_info;
	int result = 0;

	if (netif_get_info(name, urule) == 0) {
		if (strcmp(name, urule->ppp_name) != 0) {
			/* it's a ppp link */
			memset(&ppp_info, 0, sizeof(struct netif_rule_user));
			if (netif_get_info(urule->ppp_name, &ppp_info) == 0) {
				urule->ip = ppp_info.ip;
				urule->mask = ppp_info.mask;
				urule->gw[0] = ppp_info.gw[0];
			}
		}
	} 
	else {
		result = -1;
	}
	
	return result;
}

int netif_update_status(char *name, unsigned short n, unsigned short status )
{
	struct netif_rule_user urule;
	int cmd;
	int len = sizeof(struct netif_rule_user);
	int sockfd;
	int result = 0;

	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		return -1;
                		
	strncpy(urule.name, name, sizeof(urule.name)-1 );
	cmd = NETIF_SO_SET_LINK;
	urule.opcode = NETIF_OP_STATUS;
	urule.param = n;
	urule.status[n] = status;

	result = setsockopt(sockfd, IPPROTO_IP,	cmd, &urule, len);
	
	close(sockfd);	

	return result;
}

/**
 * @brief get netif_mgmt link data by given phy if name
 *
 * read netif_mgmt data from kernel module by given a phy if name
 *
 * @return 0 - no error; <0 - error
 *   
 */
int netif_get_link(
	char *name, char *pppname, uint16_t *type, 
	uint32_t *ip, uint32_t *gw1, uint32_t *gw2, uint32_t *mask, 
	uint16_t *status1, uint16_t *status2, 
	uint64_t *bw_up, uint64_t *bw_down, 
	uint32_t *testip, uint16_t *weight
)
{
	struct netif_rule_user urule;
	
	memset(&urule,0,sizeof(urule));
	
	if(netif_get_whole_info(name, &urule) != 0)
		return -1;

	if (pppname) 
		strncpy(pppname, urule.ppp_name, sizeof(urule.ppp_name) - 1);

	if (type) 
		*type = urule.type;

	if (ip) 
		*ip = urule.ip;
	
	if (gw1) 
		*gw1 = urule.gw[0];
	
	if (gw2) 
		*gw2 = urule.gw[1];
	
	if (mask) 
		*mask = urule.mask;
	
	if (status1) 
		*status1 = urule.status[0];
	
	if (status2) 
		*status2 = urule.status[1];

	if (bw_up) 
		*bw_up = urule.bw_up;

	if (bw_down) 
		*bw_down = urule.bw_down;

	if (testip) 
		*testip = urule.testip;

	if (weight) 
		*weight = urule.max_weight;
	
	return 0;
}

/**
 * @brief set netif_mgmt link data by given phy if names and opcode
 *
 * set netif_mgmt data to kernel module by given phy if name  opcode
 *
 * @return 0 - no error; <0 - error
 *   
 */
int netif_set_link(
	int op_code, char *phyifname, 			/* op_code and if name */
	char *pppname, uint16_t type,			/* NETIF_OP_LINK */
	uint16_t status1, uint16_t status2, uint16_t param,		/* NETIF_OP_STATUS */
	uint32_t gw1, uint32_t gw2,				/* NETIF_OP_GW */
	uint64_t bw_up, uint64_t bw_down,		/* NETIF_OP_BW */
	uint32_t testip,						/* NETIF_OP_CHECK */
	uint16_t weight							/* NETIF_OP_WEIGHT */
)
{
	int cmd;
	struct netif_rule_user urule;
	int len = sizeof(struct netif_rule_user);
	
	int sockfd;
	int result = 0;

	memset(&urule,0,sizeof(urule));

	memcpy(urule.name, phyifname, sizeof(urule.name));

	/* NETIF_OP_LINK */
	if (pppname)
		memcpy(urule.ppp_name, pppname, sizeof(urule.ppp_name));
	urule.type = type;

	/* NETIF_OP_STATUS */
	urule.param = param;
	urule.status[0] = status1;
	urule.status[1] = status2;

	/* NETIF_OP_GW */
	urule.gw[0] = gw1;
	urule.gw[1] = gw2;

	/* NETIF_OP_BW */
	urule.bw_up = bw_up;
	urule.bw_down = bw_down;

	/* NETIF_OP_CHECK */
	urule.testip = testip;

	/* NETIF_OP_WEIGHT */
	urule.max_weight = weight;

	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		return -1;
                		
	cmd = NETIF_SO_SET_LINK;
	urule.opcode = op_code;
	
	result = setsockopt(sockfd, IPPROTO_IP, cmd, &urule, len);
	close(sockfd);
	
	return result;
}

/**
 * @brief set netif_mgmt lb data
 *
 * set netif_mgmt data to kernel module
 *
 * @return 0 - no error; <0 - error
 *   
 */
int netif_set_lb(uint16_t operation, uint16_t lb_algo)
{
	int cmd;
	struct netif_rule_user urule;
	int len = sizeof(struct netif_rule_user);
	
	int sockfd;
	int result = 0;

	memset(&urule,0,sizeof(urule));
	urule.operation = operation;
	urule.opcode = lb_algo;

	if ((sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
		return -1;
                		
	cmd = NETIF_SO_SET_LB;
	
	result = setsockopt(sockfd, IPPROTO_IP, cmd, &urule, len);
	close(sockfd);
	
	return result;
}

